#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014-2016 The Regents of the University of California,
#               Antonino Ingargiola <tritemio@gmail.com>
#
"""

The module :mod:`select_bursts` defines functions to select
bursts according to different criteria.

These functions are usually passed to
:meth:`Data.select_bursts() <fretbursts.burstlib.Data.select_bursts>`.
For example::

    ds = d.select_bursts(select_bursts.E, th1=0.2, th2=0.6)

returns a new object `ds` containing only the bursts of `d` that pass the
specified selection criterium (`E` between 0.2 and 0.6 in this case).

"""

from __future__ import print_function, absolute_import
from builtins import range, zip

import numpy as np
from scipy import stats

from .utils.misc import clk_to_s as _clk_to_s
from .ph_sel import Ph_sel


## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
#  BURSTS SELECTION FUNCTIONS
#

def str_G(gamma, donor_ref):
    """A string indicating gamma value and convention for burst size correction.
    """
    return "Gd%.1f" % gamma if donor_ref else "Ga_%.1f" % gamma

## Selection on E or S values
def E(d, ich=0, E1=-np.inf, E2=np.inf):
    """Select bursts with E between E1 and E2."""
    assert E1 <= E2, 'Threshold E1 (%.2f) must be <= of E2 (%.2f)' % (E1, E2)
    burst_mask = (d.E[ich] >= E1)*(d.E[ich] <= E2)
    return burst_mask, ''

def S(d, ich=0, S1=-np.inf, S2=np.inf):
    assert S1 <= S2, 'Threshold S1 (%.2f) must be <= of S2 (%.2f)' % (S1, S2)
    """Select bursts with S between S1 and S2."""
    burst_mask = (d.S[ich] >= S1)*(d.S[ich] <= S2)
    return burst_mask, ''

def ES(d, ich=0, E1=-np.inf, E2=np.inf, S1=-np.inf, S2=np.inf, rect=True):
    """Select bursts with E between `E1` and `E2` and S between `S1` and `S2`.

    When `rect` is True the selection is rectangular otherwise is elliptical.

    See also:
        For plotting the ES region selected by (`E1`, `E2`, `S1`, `S2`, `rect`):

        - :func:`fretbursts.burst_plot.plot_ES_selection`
    """
    assert E1 <= E2, 'Threshold E1 (%.2f) must be <= of E2 (%.2f)' % (E1, E2)
    assert S1 <= S2, 'Threshold S1 (%.2f) must be <= of S2 (%.2f)' % (S1, S2)
    if rect:
        return ES_rect(d, ich, E1=E1, E2=E2, S1=S1, S2=S2)
    else:
        return ES_ellips(d, ich, E1=E1, E2=E2, S1=S1, S2=S2)

def ES_rect(d, ich=0, E1=-np.inf, E2=np.inf, S1=-np.inf, S2=np.inf):
    """Select bursts inside the rectangle defined by E1, E2, S1, S2.
    """
    burst_mask = (d.S[ich] >= S1)*(d.S[ich] <= S2) * \
            (d.E[ich] >= E1)*(d.E[ich] <= E2)
    return burst_mask, ''

def ES_ellips(d, ich=0, E1=-1e3, E2=1e3, S1=-1e3, S2=1e3):
    """Select bursts with E-S inside an ellipsis inscribed in E1, E2, S1, S2.
    """
    def ellips(x, y, x1, x2, y1, y2):
        rx, ry = 0.5*abs(x2-x1), 0.5*abs(y2-y1)
        return ((x - np.mean([x1,x2]))/rx)**2 + ((y - np.mean([y1,y2]))/ry)**2

    burst_mask = (ellips(d.E[ich], d.S[ich], E1, E2, S1, S2) <= 1)
    return burst_mask, ''

## Selection on static burst size, width or period
def period(d, ich=0, bp1=0, bp2=None):
    """Select bursts from period bp1 to period bp2 (included)."""
    if bp2 is None: bp2 = d.bp[ich].max()
    burst_mask = (d.bp[ich] >= bp1)*(d.bp[ich] <= bp2)
    return burst_mask, ''

def time(d, ich=0, time_s1=0, time_s2=None):
    """Select the burst starting from time_s1 to time_s2 (in seconds)."""
    burst_start = d.mburst[ich].start * d.clk_p
    if time_s2 is None: time_s2 = burst_start.max()
    burst_mask = (burst_start >= time_s1)*(burst_start <= time_s2)
    return burst_mask, ''

def nd(d, ich=0, th1=20, th2=np.inf):
    """Select bursts with (nd >= th1) and (nd <= th2)."""
    assert th1 <= th2, 'th1 (%.2f) must be <= of th2 (%.2f)' % (th1, th2)
    bursts_mask = (d.nd[ich] >= th1)*(d.nd[ich] <= th2)
    return bursts_mask, ''

def na(d, ich=0, th1=20, th2=np.inf):
    """Select bursts with (na >= th1) and (na <= th2)."""
    assert th1 <= th2, 'th1 (%.2f) must be <= of th2 (%.2f)' % (th1, th2)
    bursts_mask = (d.na[ich] >= th1)*(d.na[ich] <= th2)
    return bursts_mask, ''


def naa(d, ich=0, th1=20, th2=np.inf, gamma=1., beta=1., donor_ref=True,
        naa_comp=False, naa_aexonly=True):
    """Select bursts with (naa >= th1) and (naa <= th2).

    The `naa` quantity can be optionally corrected using gamma and beta
    factors.

    Arguments:
        th1, th2 (floats): lower (`th1`) and upper (`th2`) bounds for
            selecting `naa`. By default `th2 = inf` (i.e. no upper limit).
        gamma, beta (floats): arguments used to compute gamma- and
            beta-corrected burst sizes. See
            :meth:`fretbursts.burstlib.Data.burst_sizes_ich` for details.
        donor_ref (bool): Select the convention for `naa` correction.
            If True (default), uses `naa / (beta * gamma)`. Otherwise,
            uses `naa / beta`. It is suggested to use the same `donor_ref`
            convention when combining `Dex size` and `naa` burst selections
            so that the thresholds values of the two selections will be
            commensurable.
        na_comp (bool): **[PAX-only]** If True, multiply the `na` term
            by `(1 + Wa/Wd)`, where Wa and Wd are the D and A alternation
            durations (typically Wa/Wd = 1).
        naa_aexonly (bool): **[PAX-only]** if True, the `naa` term is
            corrected to include only A emission due to A excitation.
            If False, the `naa` term includes all the counts in DAexAem.
            The `naa` term also depends on the `naa_comp` argument.
        naa_comp (bool): **[PAX-only]** If True, multiplies the `naa` term by
            `(1 + Wa/Wd)` where Wa and Wd are the D and A alternation
            durations (typically Wa/Wd = 1). The `naa` term also depends on
            the `naa_aexonly` argument.

    See also:
        - :meth:`fretbursts.burstlib.Data.burst_sizes_pax_ich`.
    """
    assert th1 <= th2, 'th1 (%.2f) must be <= of th2 (%.2f)' % (th1, th2)
    aex_dex_ratio = d._aex_dex_ratio
    naa_term = d.naa[ich].copy()
    if 'PAX' in d.meas_type and naa_aexonly:
        naa_term -= aex_dex_ratio * d.nar[ich]
    if 'PAX' in d.meas_type and naa_comp:
        naa_term *= (1 + aex_dex_ratio)
    naa_term /= beta
    if donor_ref:
        naa_term /= gamma
    bursts_mask = (naa_term >= th1) * (naa_term <= th2)
    return bursts_mask, ''


def size(d, ich=0, th1=20, th2=np.inf, add_naa=False, gamma=1., beta=1.,
         donor_ref=True, ph_sel=None, naa_aexonly=False, naa_comp=False,
         na_comp=False):
    """Select bursts with burst sizes (i.e. counts) between `th1` and `th2`.

    The burst size is the number of photon in a burst. By default it
    includes all photons during donor excitation (`Dex`).
    To add *AexAem* photons to the burst size use `add_naa=True`.
    If `ph_sel` is specified use a PAX-specific definition of size
    as defined in :meth:`fretbursts.burstlib.Data.burst_sizes_pax_ich`.

    Arguments:
        d (Data object): the object containing the measurement.
        ich (int): the spot number, only relevant for multi-spot. In
            single-spot data there is only CH-0 so this argument may be
            omitted. Default 0.
        th1, th2 (floats): select bursts with ``th1 <= size <= th2``.
            Default `th2 = inf` (i.e. no upper limit).
        add_naa (boolean): when True, add AexAem photons when computing burst
            burst size. Default False.
        gamma, beta (floats): arguments used to compute gamma- and
            beta-corrected burst sizes. See
            :meth:`fretbursts.burstlib.Data.burst_sizes_ich` for details.
        donor_ref (bool): Select the convention for `naa` correction.
            See :meth:`fretbursts.burstlib.Data.burst_sizes_ich` for details.
        ph_sel (Ph_sel object or None): if not None, use PAX-specific
            burst size definition. ph_sel defines which terms are included
            in the burst size.
        na_comp (bool): **[PAX-only]** If True, multiply the `na` term
            by `(1 + Wa/Wd)`, where Wa and Wd are the D and A alternation
            durations (typically Wa/Wd = 1).
        naa_aexonly (bool): **[PAX-only]** if True, the `naa` term is
            corrected to include only A emission due to A excitation.
            If False, the `naa` term includes all the counts in DAexAem.
            The `naa` term also depends on the `naa_comp` argument.
        naa_comp (bool): **[PAX-only]** If True, multiply the `naa` term by
            `(1 + Wa/Wd)` where Wa and Wd are the D and A alternation
            durations (typically Wa/Wd = 1). The `naa` term also depends on
            the `naa_aexonly` argument.

    Returns:
        A tuple containing an array (the burst mask) and a string which
        briefly describes the selection.

    See also:
        - :meth:`fretbursts.burstlib.Data.burst_sizes_ich`.
        - :meth:`fretbursts.burstlib.Data.burst_sizes_pax_ich`.
    """
    assert th1 <= th2, 'th1 (%.2f) must be <= of th2 (%.2f)' % (th1, th2)
    kws = dict(ich=ich, gamma=gamma, beta=beta, donor_ref=donor_ref)
    if 'PAX' in d.meas_type and ph_sel is not None:
        kws.update(ph_sel=ph_sel, naa_aexonly=naa_aexonly, naa_comp=naa_comp,
                   na_comp=na_comp)
        burst_size = d.burst_sizes_pax_ich(**kws)
    else:
        kws.update(add_naa=add_naa)
        burst_size = d.burst_sizes_ich(**kws)

    if d.nch > 1 and (np.size(th1) == d.nch):
        th1 = th1[ich]
    if d.nch > 1 and (np.size(th2) == d.nch):
        th2 = th2[ich]
    bursts_mask = (burst_size >= th1) * (burst_size <= th2)
    s = "size_th%d" % th1
    if th2 < 1000:
        s += "_th2_%d" % th2
    return bursts_mask, s + str_G(gamma, donor_ref)

def width(d, ich=0, th1=0.5, th2=np.inf):
    """Select bursts with (width >= th1) and (width <= th2), in ms."""
    assert th1 <= th2, 'th1 (%.2f) must be <= of th2 (%.2f)' % (th1, th2)
    th1, th2 = th1*1e-3/d.clk_p, th2*1e-3/d.clk_p
    burst_width = d.mburst[ich].width
    bursts_mask = (burst_width >= th1)*(burst_width <= th2)
    return bursts_mask, ''

def sbr(d, ich=0, th1=0, th2=np.inf):
    """Select bursts with SBR between `th1` and `th2`."""
    assert th1 <= th2, 'th1 (%.2f) must be <= of th2 (%.2f)' % (th1, th2)
    if 'sbr' not in d:
        d.calc_sbr()
    sbr_ich = d.sbr[ich]
    bursts_mask = (sbr_ich >= th1)*(sbr_ich <= th2)
    return bursts_mask, ''

def peak_phrate(d, ich=0, th1=0, th2=np.inf):
    """Select bursts with peak phtotons rate between th1 and th2 (cps).

    Note that this function requires to compute the peak photon rate
    first using :meth:`fretbursts.burstlib.Data.calc_max_rate`.
    """
    rate = d.max_rate[ich]
    mask = (rate >= th1)*(rate <= th2)
    return mask, ''

def brightness(d, ich=0, th1=0, th2=np.inf, add_naa=False, gamma=1, beta=1,
               donor_ref=True):
    """Select bursts with size/width between th1 and th2 (cps).
    """
    sizes = d.burst_sizes_ich(ich=ich, gamma=gamma, add_naa=add_naa, beta=beta,
                              donor_ref=donor_ref)
    brightness = sizes/d.burst_widths[ich]
    mask = (brightness >= th1)*(brightness <= th2)
    return mask, ''


def nda_percentile(d, ich=0, q=50, low=False, gamma=1., add_naa=False):
    """Select bursts with SIZE >= q-percentile (or <= if `low` is True)

    `gamma` and `add_naa` are passed to
    :meth:`fretbursts.burstlib.Data.burst_sizes_ich` to compute the burst size.
    """
    burst_size = d.burst_sizes_ich(ich, gamma=gamma, add_naa=add_naa)
    q_percentile = np.percentile(burst_size, q=q)
    if low: bursts_mask = (burst_size <= q_percentile)
    else: bursts_mask = (burst_size >= q_percentile)
    return bursts_mask, 'perc%d' % q

def topN_nda(d, ich=0, N=500, gamma=1., add_naa=False):
    """Select the N biggest bursts in the channel.

    `gamma` and `add_naa` are passed to
    :meth:`fretbursts.burstlib.Data.burst_sizes_ich` to compute the burst size.
    """
    burst_size = d.burst_sizes_ich(ich, gamma=gamma, add_naa=add_naa)
    index_sorted = burst_size.argsort()
    burst_mask = np.zeros(burst_size.size, dtype=bool)
    burst_mask[index_sorted[-N:]] = True
    return burst_mask, 'topN%d%s' % (N, str_G(gamma, True))

def topN_max_rate(d, ich=0, N=500):
    """Select `N` bursts with the highest max burst rate.
    """
    max_burst_rate = d.max_rate[ich]
    index_sorted = max_burst_rate.argsort()
    burst_mask = np.zeros(max_burst_rate.size, dtype=bool)
    burst_mask[index_sorted[-N:]] = True
    return burst_mask, 'topN_MaxRate%d' % N

def topN_sbr(d, ich=0, N=200):
    """Select the top `N` bursts with hightest SBR."""
    if 'sbr' not in d:
        d.calc_sbr()
    sbr_ich = d.sbr[ich]
    index_sorted = sbr_ich.argsort()
    burst_mask = np.zeros(sbr_ich.size, dtype=bool)
    burst_mask[index_sorted[-N:]] = True
    return burst_mask, 'top%d_sbr' % N



## Selection on burst time (nearby, overlapping or isolated bursts)
def single(d, ich=0, th=1):
    """Select bursts that are at least th millisec apart from the others."""
    th = th*1e-3/d.clk_p
    burst_start = d.mburst[ich].start
    burst_stop = d.mburst[ich].stop
    gap_mask = (burst_start[1:] - burst_stop[:-1]) >= th
    bursts_mask = np.hstack([gap_mask,False])*np.hstack([False,gap_mask])
    return bursts_mask, ''

def consecutive(d, ich=0, th1=0, th2=np.inf, kind='both'):
    """Select consecutive bursts with th1 <= separation <= th2 (in sec.).

    Arguments:
        kind (string): valid values are 'first' to select the first burst
            of each pair, 'second' to select the second burst of each pair
            and 'both' to select both bursts in each pair.
    """
    assert th1 <= th2, 'th1 (%.2f) must be <= of th2 (%.2f)' % (th1, th2)
    kind_valids = ['first', 'second', 'both']
    assert kind in kind_valids, \
        "Invalid value for 'kind'. Valid values are %s" % str(kind_valids)

    bseparation = d.mburst[ich].separation*d.clk_p
    pair_mask = (bseparation >= th1)*(bseparation <= th2)

    bursts_mask = np.zeros(d.num_bursts[ich], dtype=bool)
    if kind in ['first', 'both']:
        bursts_mask += np.hstack([pair_mask, (False,)])
    if kind in ['second', 'both']:
        bursts_mask += np.hstack([(False,), pair_mask])
    return bursts_mask, ''

## Selection on burst size vs BG
def nd_bg(d, ich=0, F=5):
    """Select bursts with (nd >= bg_dd*F)."""
    bg_burst = d.bg_dd[ich][d.bp[ich]] * d.mburst[ich].width * d.clk_p
    bursts_mask = (d.nd[ich] >= F*bg_burst)
    return bursts_mask, ''

def na_bg(d, ich=0, F=5):
    """Select bursts with (na >= bg_ad*F)."""
    bg_burst = d.bg_ad[ich][d.bp[ich]] * d.mburst[ich].width * d.clk_p
    bursts_mask = (d.na[ich] >= F*bg_burst)
    return bursts_mask, ''

def naa_bg(d, ich=0, F=5):
    """Select bursts with (naa >= bg_aa*F)."""
    bg_burst = d.bg_aa[ich][d.bp[ich]] * d.mburst[ich].width * d.clk_p
    bursts_mask = (d.naa[ich] >= F*bg_burst)
    return bursts_mask, ''

def nt_bg(d, ich=0, F=5):
    """Select bursts with (nt >= bg*F)."""
    bg_burst = d.bg[ich][d.bp[ich]] * d.mburst[ich].width * d.clk_p
    bursts_mask = (d.nt[ich] > F*bg_burst)
    return bursts_mask, ''

## Selection on burst size vs BG (probabilistic)
def na_bg_p(d, ich=0, P=0.05, F=1.):
    """Select bursts w/ AD signal using P{F*BG>=na} < P."""
    accept_ch_bg_rate = d.bg_mean(Ph_sel(Dex='Aem'))[ich]
    bursts_width = _clk_to_s(d.mburst[ich].width)
    max_num_bg_ph = stats.poisson(F*accept_ch_bg_rate*bursts_width).isf(P)
    #print("Min num. ph = ",  max_num_bg_ph)
    bursts_mask = (d.na[ich] >= max_num_bg_ph)
    return bursts_mask, ''

def nd_bg_p(d, ich=0, P=0.05, F=1.):
    """Select bursts w/ DD signal using P{F*BG>=nd} < P."""
    donor_ch_bg_rate = d.bg_mean(Ph_sel(Dex='Dem'))[ich]
    bursts_width = _clk_to_s(d.mburst[ich].width)
    max_num_bg_ph = stats.poisson(F*donor_ch_bg_rate*bursts_width).isf(P)
    #print("Min num. ph = ", max_num_bg_ph)
    bursts_mask = (d.nd[ich] >= max_num_bg_ph)
    return bursts_mask, ''

def naa_bg_p(d, ich=0, P=0.05, F=1.):
    """Select bursts w/ AA signal using P{F*BG>=naa} < P."""
    A_em_ex_bg_rate = d.bg_mean(Ph_sel(Aex='Aem'))[ich]
    bursts_width = _clk_to_s(d.mburst[ich].width)
    max_num_bg_ph = stats.poisson(F*A_em_ex_bg_rate*bursts_width).isf(P)
    #print("Min num. ph = ", max_num_bg_ph)
    bursts_mask = (d.naa[ich] >= max_num_bg_ph)
    return bursts_mask, ''

def nt_bg_p(d, ich=0, P=0.05, F=1.):
    """Select bursts w/ signal using P{F*BG>=nt} < P."""
    bg_rate = d.rate_m[ich]
    bursts_width = _clk_to_s(d.mburst[ich].width)
    max_num_bg_ph = stats.poisson(F*bg_rate*bursts_width).isf(P)
    #print("Min num. ph = ", max_num_bg_ph)
    #print("burst width (ms) = ", bursts_width*1e3)
    #print("Poisson rate = ", bg_rate*bursts_width)
    #print("rate = ", bg_rate)
    bursts_mask = (d.nt[ich] >= max_num_bg_ph)
    return bursts_mask, ''



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014 Antonino Ingargiola <tritemio@gmail.com>
#
"""
SPC Format (Beker & Hickl)
--------------------------

48-bit element in little endian (<) format

Drawing (note: each char represents 2 bits)::

    bit: 64        48                          0
         0000 0000 XXXX XXXX XXXX XXXX XXXX XXXX
                   '-------' '--' '--'   '-----'
    field names:       a      c    b        d

         0000 0000 XXXX XXXX XXXX XXXX XXXX XXXX
                   '-------' '--' '--' '-------'
    numpy dtype:       a      c    b    field0

    macrotime = [ b  ] [     a     ]  (24 bit)
    detector  = [ c  ]                (8 bit)
    nanotime  = [  d  ]               (12 bit)

    overflow bit: 13, bit_mask = 2^(13-1) = 4096
"""

from builtins import range, zip

import numpy as np


def load_spc(fname):
    """Load data from Becker&Hickl SPC files.

    Returns:
        3 numpy arrays: timestamps, detector, nanotime
    """
    spc_dtype = np.dtype([('field0', '<u2'), ('b', '<u1'), ('c', '<u1'),
                          ('a', '<u2')])
    data = np.fromfile(fname, dtype=spc_dtype)

    nanotime =  4095 - np.bitwise_and(data['field0'], 0x0FFF)
    detector = data['c']

    # Build the macrotime (timestamps) using in-place operation for efficiency
    timestamps = data['b'].astype('int64')
    np.left_shift(timestamps, 16, out=timestamps)
    timestamps += data['a']

    # extract the 13-th bit from data['field0']
    overflow = np.bitwise_and(np.right_shift(data['field0'], 13), 1)
    overflow = np.cumsum(overflow, dtype='int64')

    # Add the overflow bits
    timestamps += np.left_shift(overflow, 24)

    return timestamps, detector, nanotime



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014-2016 The Regents of the University of California,
#               Antonino Ingargiola <tritemio@gmail.com>
#
"""
The module :mod:`phtools.burstsearch` provides the low-level (or core)
burst search and photon counting functions.
This module also provides :class:`Bursts`, a container for a set of bursts.
:class:`Bursts` provides attributes for the main burst quatitites (`istart`,
`istop`, `start`, `stop`, `counts`,  `width`, etc...). It implements the
iterator interface (iterate burst by burst). Moreover :class:`Bursts` can
be indexed (`[]`, i.e. `getitem` interface) supporting the same indexing as a
numpy 1-D array.

The burst search functions return a 2-D array (burst array) of shape Nx4,
where N is the number of bursts. This array can used to build a `Bursts` object
using::

    Bursts(bursts_array)

As an example, let assume having a burst array `bursts`. To take a slice of
only the first 10 bursts you can do::

    bursts10 = bursts[:10]   # new Bursts object with the first 10 bursts

To obtain the burst start of all the bursts::

    bursts.start

To obtain the burst counts (number of photons) for the 10-th to 20-th burst::

    bursts[10:20].counts

For efficiency, when iterating over `Bursts` the returned burst is a
named tuple :class:`Burst`, which implements the same attributes as `Bursts`
(istart, istop, start, stop, counts and width).
This results in faster iteration and attribute access than using `Bursts`
objects with only one burst.

Three methods allow to transform Bursts to refer to a new timestamps array:

- :meth:`Bursts.recompute_times`
- :meth:`Bursts.recompute_index_expand`
- :meth:`Bursts.recompute_index_reduce`

Finally, in order to support fusion of consecutive bursts, we provide the class
:class:`BurstsGap` (and single-burst version :class:`BurstGap`) which add the
attributes `gap` and `gap_counts` that contains the duration and the number
of photons in gaps inside a burst. The attribute `width` is the total burst
duration minus `gap`, while `counts` is the total number of photons minus
photons falling inside gaps (gaps are open intervals, do not include edges).
"""

from __future__ import division, print_function
from builtins import range, zip

from collections import namedtuple
import numpy as np
import pandas as pd

from fretbursts.utils.misc import pprint

pd.set_option('display.max_rows', 10)


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
#  LOW-LEVEL BURST SEARCH FUNCTIONS
#

def bsearch_py(times, L, m, T, slice_=None,
               label='Burst search', verbose=True):
    """Sliding window burst search. Pure python implementation.

    Finds bursts in the array `time` (int64). A burst starts when the photon rate
    is above a minimum threshold, and ends when the rate falls below the same
    threshold. The rate-threshold is defined by the ratio `m`/`T` (`m` photons
    in a time interval `T`). A burst is discarded if it has less than `L`
    photons.

    Arguments:
        times (array, int64): array of timestamps on which to perform the search
        L (int): minimum number of photons in a bursts. Bursts with size
            (or counts) < L are discarded.
        m (int): number of consecutive photons used to compute the rate.
        T (float): max time separation of `m` photons to be inside a burst
        slice_ (tuple): 2-element tuple used to slice times
        label (string): a label printed when the function is called
        verbose (bool): if False, the function does not print anything.

    Returns:
        Array of burst data Nx4, type int64.
        Column order is: istart, istop, start, stop.
    """
    if verbose:
        pprint('Python search (v): %s\n' % label)

    i_time0 = 0
    if slice_ is not None:
        times = times[slice_[0]:slice_[1]]
        i_time0 = slice_[0]

    bursts = []
    above_min_rate = (times[m-1:] - times[:times.size-m+1]) <= T

    it = enumerate(above_min_rate)
    for i, above_min_rate_ in it:
        if not above_min_rate_:
            continue

        i_start = i
        for i, above_min_rate_ in it:
            if not above_min_rate_:
                break
        i_stop = i + m - 2  # index of last ph in burst
        if i_stop - i_start + 1 >= L:
            bursts.append((i_start, i_stop, times[i_start], times[i_stop]))

    if above_min_rate_:
        # Correct burst-stop off by 1 when last burst does not finish
        i_stop += 1
        if i_stop - i_start + 1 >= L:
            bursts.pop()
            bursts.append((i_start, i_stop, times[i_start], times[i_stop]))

    bursts = np.array(bursts, dtype='int64')
    if bursts.size > 0:
        # Add slice offset to istart and istop
        bursts[:, :2] += i_time0
    return bursts


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
#  Functions to count D and A photons in bursts
#

def count_ph_in_bursts(bursts, mask):
    """Counts number of photons in each burst counting only photons in `mask`.

    This function takes a :class:`Bursts` object and a boolean mask (photon
    selection) and computes the number of photons selected by the mask.
    It is used, for example, to count donor and acceptor photons
    in each burst.

    For a multi-channel version see :func:`mch_count_ph_in_bursts_py`.

    Arguments:
        bursts (Bursts object): the bursts used as input
        mask (1D boolean array): the photon mask. The boolean mask must be
            of the same size of the timestamp array used for burst search.

    Returns:
        A 1D array containing the number of photons in each burst
        counting only photons in the selection mask.
    """
    num_ph = np.zeros(bursts.num_bursts, dtype=np.int32)
    for i, burst in enumerate(bursts):
        # Counts photons between start and end of current `burst`
        num_ph[i] = mask[burst.istart : burst.istop + 1].sum()
    return num_ph


def mch_count_ph_in_bursts_py(Mburst, Mask):
    """Counts number of photons in each burst counting only photons in `Mask`.

    This multi-channel function takes a list of a :class:`Bursts` objects and
    photon masks and computes the number of photons selected by the mask
    in each channel.

    It is used, for example, to count donor and acceptor photons in
    each burst.

    For a single-channel version see :func:`count_ph_in_bursts_py`.

    Arguments:
        Mburst (list Bursts objects): a list of bursts collections, one per ch.
        Mask (list of 1D boolean arrays): a list of photon masks (one per ch),
            For each channel, the boolean mask must be of the same size of the
            timestamp array used for burst search.

    Returns:
        A list of 1D array, each containing the number of photons
        in each burst counting only photons in the selection mask.
    """
    return [np.asfarray(count_ph_in_bursts(bursts, mask))
            for bursts, mask in zip(Mburst, Mask)]


##
#  Try to import the optimized Cython functions
#

try:
    from burstsearch_c import bsearch_c
    bsearch = bsearch_c
    print(" - Optimized (cython) burst search loaded.")
except ImportError:
    bsearch = bsearch_py
    print(" - Fallback to pure python burst search.")

try:
    from burstsearch_c import mch_count_ph_in_bursts_c
    mch_count_ph_in_bursts = mch_count_ph_in_bursts_c
    print(" - Optimized (cython) photon counting loaded.")
except ImportError:
    mch_count_ph_in_bursts = mch_count_ph_in_bursts_py
    print(" - Fallback to pure python photon counting.")


class Burst(namedtuple('Burst', ['istart', 'istop', 'start', 'stop'])):
    """Container for a single burst."""
    __slots__ = ()

    @property
    def width(self):
        """Burst duration in timestamps unit."""
        return self.stop - self.start

    @property
    def counts(self):
        """Number of photons in the burst."""
        return self.istop - self.istart + 1

    @property
    def ph_rate(self):
        """Photon rate in the burst (total photon counts/duration)."""
        return self.counts / self.width


class BurstGap(namedtuple('BurstGap',
                          'istart istop start stop gap gap_counts')):
    __slots__ = ()

    @staticmethod
    def from_burst(burst):
        """Build a `BurstGap` from a :class:`Burst` object."""
        return BurstGap(start=burst.start, stop=burst.stop,
                        istart=burst.istart, istop=burst.istop,
                        gap=0, gap_counts=0)

    @property
    def width(self):
        """Burst duration in timestamps unit, minus `gap` time."""
        return self.stop - self.start - self.gap

    @property
    def counts(self):
        """Number of photons in the burst, minus `gap_counts`."""
        return self.istop - self.istart + 1 - self.gap_counts


class Bursts(object):
    """A container for burst data.

    This class provides a container for burst data. It has a
    set of attributes (`start`, `stop`, `istart`, `istop`, `counts`, `width`,
    `ph_rate`, `separation`) that can be accessed to obtain burst data.
    Only a few fundamental attributes are stored, the others are comuputed
    on-fly using python properties.

    Other attributes are `dataframe` (a `pandas.DataFrame` with the complete
    burst data), `num_bursts` (the number of bursts).

    `Bursts` objects can be built from a list of single :class:`Burst` objects
    by using the method :meth:`Bursts.from_list`, or from 2D arrays
    containing bursts data (one row per burst; columns: istart, istop, start,
    stop) such as the ones returned by burst search functions (e.g.
    :func:`bsearch_py`).

    `Bursts` objects are iterable, yielding one burst a time (:class:`Burst`
    objects). `Bursts` can be compared for equality (with `==`) and copied
    (:meth:`Bursts.copy`).

    Additionally basic methods for burst manipulation are provided:

    - `recompute_times` recompute start and stop times using the current
      start and stop index and a new timestamps array passed as argument.
    - `recompute_index_*` recompute start and stop indexes to refer to an
      expanded or reduced timestamp selection.

    Other methods are:

    - `and_gate` computing burst intersection with a second set of bursts.
      Used to implement the dual-channel burst search (DCBS).

    Methods that may be implemented in the future:

    - `or_gate`: computing union with a second set of bursts.
    - `fuse_bursts`: fuse nearby bursts.

    """
    _i_istart, _i_istop, _i_start, _i_stop = 0, 1, 2, 3
    _ncols = 4

    def __init__(self, burstarray):
        if burstarray.ndim == 1:
            assert burstarray.size == self._ncols
            burstarray = burstarray[np.newaxis, :]
        self.data = burstarray

    @classmethod
    def empty(cls, num_bursts=0):
        """Return an empty `Bursts()` object."""
        return cls(np.zeros((num_bursts, cls._ncols),
                            dtype=np.int64))

    @classmethod
    def from_list(cls, bursts_list):
        """Build a new `Bursts()` object from a list of :class:`Burst`.
        """
        bursts = cls.empty(len(bursts_list))
        for i, burst in enumerate(bursts_list):
            bursts.istart[i], bursts.istop[i] = burst.istart, burst.istop
            bursts.start[i], bursts.stop[i] = burst.start, burst.stop
        return bursts

    @classmethod
    def merge(cls, list_of_bursts, sort=False):
        """Merge `Bursts` in `list_of_bursts`, returning a new `Bursts` object.
        """
        mergedata = np.vstack([b.data for b in list_of_bursts])
        if sort:
            # Sort by start times, and when equal by stop times
            indexsort = np.lexsort((mergedata[:, 3], mergedata[:, 2]))
            mergedata = mergedata[indexsort]
        return cls(mergedata)

    #
    # Basic interface
    #
    def copy(self):
        """Return a new copy of current `Bursts` object."""
        return self.__class__(self.data.copy())

    def __getitem__(self, i):
        return self.__class__(self.data[i])

    def __iter__(self):
        for bdata in self.data:
            # Bursts has a fast __init__() but relatively slow
            # attribute access:
            #yield Bursts(bdata)

            # Burst (namedtuple) has fast attribute access (>10x faster)
            # and __init__ as fast as Bursts.__init__ when avoiding the *bdata
            # syntax (which creates an additional tuple).
            #yield Burst(*bdata)
            yield Burst(bdata[0], bdata[1], bdata[2], bdata[3])

    def __eq__(self, other_bursts):
        return np.all(self.data == other_bursts.data)

    @property
    def num_bursts(self):
        """Number of bursts."""
        return self.data.shape[0]

    @property
    def size(self):
        """Number of bursts. Used for compatibility with ndarray.size.
        Use `Bursts.num_bursts` preferentially.
        """
        return self.num_bursts

    @property
    def dataframe(self):
        """A `pandas.DataFrame` containing burst data, one row per burst.
        """
        return pd.DataFrame(self.data,
                            columns=['istart', 'istop', 'start', 'stop'])

    def __repr__(self):
        return self.dataframe.__repr__()

    def _repr_html_(self):
        return self.dataframe._repr_html_()

    def join(self, bursts, sort=False):
        """Join the current `Bursts` object with another one. Returns a copy.
        """
        return self.merge([self, bursts], sort=sort)

    #
    # Burst data attributes/properties
    #
    def _set_data(self, column, values):
        """This method allows subclasses to easily extend the setters.
        """
        self.data[:, column] = values

    @property
    def start(self):
        """Time of 1st ph in each burst"""
        return self.data[:, Bursts._i_start]

    @start.setter
    def start(self, value):
        self._set_data(Bursts._i_start, value)

    @property
    def stop(self):
        """Time of last ph in each burst"""
        return self.data[:, Bursts._i_stop]

    @stop.setter
    def stop(self, value):
        self._set_data(Bursts._i_stop, value)

    @property
    def istart(self):
        """Index of 1st ph in each burst"""
        return self.data[:, Bursts._i_istart]

    @istart.setter
    def istart(self, value):
        self._set_data(Bursts._i_istart, value)

    @property
    def istop(self):
        """Index of last ph in each burst"""
        return self.data[:, Bursts._i_istop]

    @istop.setter
    def istop(self, value):
        self._set_data(Bursts._i_istop, value)

    @property
    def width(self):
        """Burst duration in timestamps units."""
        return self.stop - self.start

    @property
    def counts(self):
        """Number of photons in each burst."""
        return self.istop - self.istart + 1

    @property
    def ph_rate(self):
        """Photon rate in burst (tot size/duration)"""
        return self.counts / self.width

    @property
    def separation(self):
        """Separation between nearby bursts"""
        return self.start[1:] - self.stop[:-1]

    #
    # Burst manipulation methods
    #
    def recompute_times(self, times, out=None):
        """Recomputes start, stop times using timestamps from a new array.

        This method computes burst start, stop using the index of timestamps
        from the current object and timestamps from the passed array `times`.

        This is useful, for example, when burst search is computed on a
        "compacted" timestamps array (i.e. removing the gaps outside the
        alternation period in usALEX experiments), and afterwards the "real"
        start and stop times needs to be recomputed.

        Arguments:
            times (array): array of photon timestamps
            out (None or Bursts): if None (default), do computations on a copy
                of the current object. Otherwise, modify the `Bursts` object
                passed (can be used for in-place operations).

        Returns:
            `Bursts` object with recomputed start/stop times.
        """
        if out is None:
            out = self.copy()
        out.start = times[self.istart]
        out.stop = times[self.istop]
        return out

    def recompute_index_expand(self, mask, out=None):
        """Recompute istart and istop from selection `mask` to full timestamps.

        This method returns a new Bursts object with recomputed istart and
        istop. Old istart, istop are assumed to be index of a reduced array
        `timestamps[mask]`. New istart, istop are computed to be index of
        a "full" timestamps array of size `mask.size`.

        This is useful when performing burst search on a timestamps selection
        and we want to transform the burst data to use the index of the "full"
        timestamps array.

        Arguments:
            mask (bool array): boolean mask defining the timestamps selection
                on which the old istart and istop were computed.
            out (None or Bursts): if None (default), do computations on a copy
                of the current object. Otherwise, modify the `Bursts` object
                passed (can be used for in-place operations).

        Returns:
            `Bursts` object with recomputed istart/istop.
        """
        if out is None:
            out = self.copy()
        index = np.arange(mask.size, dtype=np.int32)
        out.istart = index[mask][self.istart]
        out.istop = index[mask][self.istop]
        return out

#    def recompute_index_reduce2(self, times_reduced, out=None):
#        """Recompute istart and istop on reduced timestamps `times_reduced`.
#
#        Extremely inefficient (but very simple!) version of
#        `Bursts.recompute_index` used for testing.
#        """
#        if out is None:
#            out = self.copy()
#
#        for i, burst in enumerate(self):
#            # The first index ([0]) accesses the tuple returned by nonzero.
#            # The second index ([0] or [-1]) accesses the array inside the
#            # tuple. This array can have size > 1 when burst start or stop
#            # happens on a repeated timestamp.
#            out[i].istart = np.nonzero(times_reduced == burst.start)[0][0]
#            out[i].istop = np.nonzero(times_reduced == burst.stop)[0][-1]
#        return out

    def recompute_index_reduce(self, times_reduced, out=None):
        """Recompute istart and istop on reduced timestamps `times_reduced`.

        This method returns a new Bursts object with same start and stop times
        and recomputed istart and istop. Old istart, istop are assumed to
        be index of a "full" timestamps array of size `mask.size`. New istart,
        istop are computed to be index of the reduced timestamps array
        `timestamps_reduced`.

        Note: it is required that all the start and stop times are
        also contained in the reduced timestamps selection.

        This method is the inverse of :meth:`recompute_index_expand`.

        Arguments:
            times_reduced (array): array of selected timestamps used to
                compute the new istart and istop. This array needs to be
                a sub-set of the original timestamps array.
            out (None or Bursts): if None (default), do computations on a copy
                of the current object. Otherwise, modify the `Bursts` object
                passed (can be used for in-place operations).

        Returns:
            `Bursts` object with recomputed istart/istop times.
        """
        if out is None:
            out = self.copy()

        # Go through the timestamps searching for start
        # and stop of each burst in order
        it = 0
        for ib, burst in enumerate(self):
            while times_reduced[it] != burst.start:
                it += 1
            out[ib].istart = it
            it_saved = it + 1

            while times_reduced[it] != burst.stop:
                it += 1
            # in case of repeated timestamps, istop needs to point
            # to the last of the repeats
            while it < times_reduced.size and times_reduced[it] == burst.stop:
                it += 1
            out[ib].istop = it - 1

            # Done with current burst, before starting a new burst,
            # reset `it` to `istart+1`
            it = it_saved
        return out

    def and_gate(self, bursts2):
        """From 2 burst arrays return bursts defined as intersection (AND rule).

        The two input burst-arrays come from 2 different burst searches.
        Returns new bursts representing the overlapping bursts in the 2 inputs
        with start and stop defined as intersection (or AND) operator.

        Both input and output are Bursts objects.

        Arguments:
            bursts_a (Bursts object): second set of bursts to be intersected
                with bursts in self. The number of bursts in self and
                `bursts_a` can be different.

        Returns:
            Bursts object containing intersections (AND) of overlapping bursts.
        """
        start_d, stop_d = self.start, self.stop
        start_a, stop_a = bursts2.start, bursts2.stop

        bursts = []
        i_d, i_a = 0, 0
        while i_d < self.num_bursts and i_a < bursts2.num_bursts:
            # Skip any disjoint burst
            if stop_a[i_a] < start_d[i_d]:
                i_a += 1
                continue
            if stop_d[i_d] < start_a[i_a]:
                i_d += 1
                continue

            # Assign start and stop according the AND rule
            if start_a[i_a] < start_d[i_d] < stop_a[i_a]:
                start_burst = self[i_d]
            elif start_d[i_d] < start_a[i_a] < stop_d[i_d]:
                start_burst = bursts2[i_a]

            if stop_d[i_d] < stop_a[i_a]:
                end_burst = self[i_d]
                i_d += 1
            else:
                end_burst = bursts2[i_a]
                i_a += 1

            burst = start_burst.copy()
            burst.stop = end_burst.stop
            burst.istop = end_burst.istop

            bursts.append(burst)

        return Bursts.from_list(bursts)


class BurstsGap(Bursts):
    """A container for bursts with optional gaps.

    This class extend Bursts adding the attributes/properties `gap`
    (a duration) and `gap_counts` (counts in gap) that allow accounting
    for gaps inside bursts.

    """
    _i_gap, _i_gap_counts = 4, 5
    _ncols = 6

    def __init__(self, burstarray):
        if burstarray.ndim == 2 and burstarray.shape[1] == Bursts._ncols:
            # Add missing gap columns
            datag = np.zeros((burstarray.shape[0], self._ncols),
                             dtype=np.int64)
            datag[:, :Bursts._ncols] = burstarray
            burstarray = datag
        super(BurstsGap, self).__init__(burstarray)

    @classmethod
    def from_list(cls, bursts_list):
        """Build a new `BurstsGap()` from a list of :class:`BurstGap`.
        """
        bursts = super(cls, cls).from_list(bursts_list)
        for i, burst in enumerate(bursts_list):
            bursts.gap[i], bursts.gap_counts[i] = burst.gap, burst.gap_counts
        return bursts

    def __iter__(self):
        for bdata in self.data:
           yield BurstGap(bdata[0], bdata[1], bdata[2], bdata[3],
                          bdata[4], bdata[5])

    @property
    def gap(self):
        """Time gap inside a burst"""
        return self.data[:, BurstsGap._i_gap]

    @gap.setter
    def gap(self, value):
        self._set_data(BurstsGap._i_gap, value)

    @property
    def gap_counts(self):
        """Number of photons falling inside gaps of each burst."""
        return self.data[:, BurstsGap._i_gap_counts]

    @gap_counts.setter
    def gap_counts(self, value):
        self._set_data(BurstsGap._i_gap_counts, value)

    @property
    def width(self):
        """Burst duration in timestamps units, minus the `gap` time.
        """
        return self.stop - self.start - self.gap

    @property
    def counts(self):
        """Number of photons in each burst, minus the `gap_counts`.
        """
        return self.istop - self.istart + 1 - self.gap_counts



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014-2016 The Regents of the University of California,
#               Antonino Ingargiola <tritemio@gmail.com>
#
"""
This model provides a class for fitting multi-channel data
(:class:`MultiFitter`) and a series of predefined functions for common
models used to fit E or S histograms.
"""

from __future__ import division, print_function, absolute_import
from builtins import range, zip

import numpy as np
import pandas as pd
import lmfit
try:
    import matplotlib.pyplot as plt
    from mpl_toolkits.axes_grid1 import make_axes_locatable
except ImportError:
    has_matplotlib = False
else:
    has_matplotlib = True

from .fit import gaussian_fitting as gf


## Utility functions
def find_max(x, y, xmin=None, xmax=None):
    """Find peak position of a curve (x, y) between `xmin` and `xmax`.
    """
    if xmin is None:
        xmin = x.min()
    if xmax is None:
        xmax = x.max()

    mask = np.where((x >= xmin)*(x <= xmax))
    return x[mask][y[mask].argmax()]


## Base function shapes
def gaussian(x, center, sigma, amplitude=1):
    """A gaussian function. Parameters: `center`, `sigma`, `amplitude`.
    """
    return amplitude * np.exp(-(x - center)**2/(2*sigma**2))

def asym_gaussian(x, center, sigma1, sigma2, amplitude):
    """A asymmetric gaussian function composed by two gaussian halves.

    This function is composed from two gaussians joined at their peak, so that
    the left and right side decay with different sigmas.

    Arguments:
        x (array): 1-D array for the independent variable
        center (float): function peak position
        sigma1 (float): sigma of the left-side gaussian (for x < center)
        sigma2 (float): sigma of the right-side gaussian (for x > center)
        amplitude (float): maximum value reach for x = center.

    Return:
        An array (same shape as `x`) with the function values.
    """

    left_half = x < center
    sigma = np.zeros_like(x)
    sigma[left_half] = sigma1
    sigma[-left_half] = sigma2
    u = (x - center)/sigma
    return amplitude*np.exp(-u*u)

def bridge_function(x, center1, center2, sigma1, sigma2, amplitude):
    """A "bridge" function, complementary of two gaussian peaks.

    Let `g` be a Gaussian function (with amplitude = 1), the bridge function
    is defined as::

        amplitude * (1 - g(x, center1, sigma1) - g(x, center2, sigma2))

    for `center1 < x < center2`. The function is 0 otherwise.

    Arguments:
        x (array): 1-D array for the independent variable
        center1 (float): center of the first gaussian (left side)
        center2 (float): center of the second gaussian (right side)
        sigma1 (float): sigma of the left-side gaussian
        sigma2 (float): sigma of the right-side gaussian
        amplitude (float): maximum (asymptotic) value of the bridge (plateau)

    Return:
        An array (same shape as `x`) with the function values.
    """
    assert center1 <= center2
    assert amplitude >= 0
    mask = (x > center1)*(x < center2)
    y = np.zeros_like(x)
    y[mask] = 1.
    y[mask] -= gaussian(x[mask], center1, sigma1)
    y[mask] -= gaussian(x[mask], center2, sigma2)
    y *= amplitude
    return y

def bridge_function2(x, center1, center2, sigma1, sigma2, amplitude):
    """A "bridge" function, geometric complementary of two gaussian peaks.

    Let `g` be a Gaussian function (with amplitude = 1), the bridge function
    is defined as:

        amplitude * (1 - g(x, center1, sigma1)) * (1 - g(x, center2, sigma2))

    for center1 < x < center2. The function is 0 otherwise.

    Arguments:
        x (array): 1-D array for the independent variable
        center1 (float): center of the first gaussian (left side)
        center2 (float): center of the second gaussian (right side)
        sigma1 (float): sigma of the left-side gaussian
        sigma2 (float): sigma of the right-side gaussian
        amplitude (float): maximum (asymptotic) value of the bridge (plateau)

    Return:
        An array (same shape as `x`) with the function values.
    """
    assert center1 <= center2
    assert amplitude >= 0
    mask = (x > center1)*(x < center2)
    y = np.zeros_like(x)
    y[mask] = amplitude
    y[mask] *= 1 - gaussian(x[mask], center1, sigma1)
    y[mask] *= 1 - gaussian(x[mask], center2, sigma2)
    y *= amplitude
    return y


##
# Factory functions that return initialized `lmfit.Model` objects
#
def factory_gaussian(center=0.1, sigma=0.1, amplitude=1):
    """Return an lmfit Gaussian model that can be used to fit data.

    Arguments are initial values for the model parameters.

    Returns
        An `lmfit.Model` object with all the parameters already initialized.
    """
    model = lmfit.models.GaussianModel()
    model.set_param_hint('center', value=center, min=-1, max=2)
    model.set_param_hint('sigma', value=sigma)
    model.set_param_hint('amplitude', value=amplitude)
    return model

def factory_asym_gaussian(center=0.1, sigma1=0.1, sigma2=0.1, amplitude=1):
    """Return a lmfit Asymmetric Gaussian model that can be used to fit data.

    For the definition of asymmetric Gaussian see :func:`asym_gaussian`.
    Arguments are initial values for the model parameters.

    Returns
        An `lmfit.Model` object with all the parameters already initialized.
    """
    model = lmfit.model.Model(asym_gaussian)
    model.set_param_hint('center', value=center, min=-1, max=2)
    model.set_param_hint('sigma1', value=sigma1)
    model.set_param_hint('sigma2', value=sigma2)
    model.set_param_hint('amplitude', value=amplitude)
    return model

def factory_two_gaussians(add_bridge=False, p1_center=0.1, p2_center=0.9,
                          p1_sigma=0.03, p2_sigma=0.03):
    """Return a 2-Gaussian + (optional) bridge model that can fit data.

    The optional "bridge" component (i.e. a plateau between the two peaks)
    is a function that is non-zero only between `p1_center` and `p2_center`
    and is defined as::

        br_amplitude * (1 - g(x, p1_center, p1_sigma) - g(x, p1_center, p2_sigma))

    where `g` is a gaussian function with amplitude = 1 and `br_amplitude`
    is the height of the plateau and the only additional parameter introduced
    by the bridge. Note that both centers and sigmas parameters in the bridge
    are the same ones of the adjacent Gaussian peaks. Therefore a
    2-Gaussian + bridge model has 7 free parameters: 3 for each Gaussian and
    an additional one for the bridge.
    The bridge function is implemented in :func:`bridge_function`.

    Arguments:
        p1_center, p2_center (float): initial values for the centers of the
            two Gaussian components.
        p1_sigma, p2_sigma (float): initial values for the sigmas of the
            two Gaussian components.
        add_bridge (bool): if True adds a bridge function between the two
            gaussian peaks. If False the model has only two Gaussians.

    Returns
        An `lmfit.Model` object with all the parameters already initialized.
    """
    peak1 = lmfit.models.GaussianModel(prefix='p1_')
    peak2 = lmfit.models.GaussianModel(prefix='p2_')
    model = peak1 + peak2

    model.set_param_hint('p1_center', value=p1_center, min=-1, max=2)
    model.set_param_hint('p2_center', value=p2_center, min=-1, max=2)
    model.set_param_hint('p1_sigma', value=p1_sigma, min=0.01, max=0.2)
    model.set_param_hint('p2_sigma', value=p2_sigma, min=0.01, max=0.2)
    model.set_param_hint('p1_amplitude', value=1, min=0.01)
    model.set_param_hint('p2_amplitude', value=1, min=0.01)
    name = '2-gaussians'

    if add_bridge:
        bridge = lmfit.model.Model(bridge_function, prefix='br_')
        model += bridge
        model.set_param_hint('br_amplitude', value=0.0001, min=0)
        model.set_param_hint('br_center1', min=0, expr='p1_center')
        model.set_param_hint('br_center2', min=0, expr='p2_center')
        model.set_param_hint('br_sigma1', min=0, expr='p1_sigma')
        model.set_param_hint('br_sigma2', min=0, expr='p2_sigma')
        name += '-bridge'
    model.name = name
    return model

def factory_three_gaussians(p1_center=0., p2_center=0.5, p3_center=1,
                            sigma=0.05):
    """Return a 3-Gaussian model that can fit data.

    The other arguments are initial values for the `center` for each
    Gaussian component plus an single `sigma` argument that is used
    as initial sigma for all the Gaussians. Note that during the fitting
    the sigma of each Gaussian is varied independently.

    Returns
        An `lmfit.Model` object with all the parameters already initialized.
    """
    peak1 = lmfit.models.GaussianModel(prefix='p1_')
    peak2 = lmfit.models.GaussianModel(prefix='p2_')
    peak3 = lmfit.models.GaussianModel(prefix='p3_')
    model = peak1 + peak2 + peak3

    model.set_param_hint('p1_center', value=p1_center, min=0, max=1)
    model.set_param_hint('p2_center', value=p2_center, min=0, max=1)
    model.set_param_hint('p3_center', value=p3_center, min=0, max=1)
    model.set_param_hint('p1_sigma', value=sigma, min=0.02, max=0.2)
    model.set_param_hint('p2_sigma', value=sigma, min=0.02, max=0.2)
    model.set_param_hint('p3_sigma', value=sigma, min=0.02, max=0.2)
    model.set_param_hint('p1_amplitude', value=1, min=0.01)
    model.set_param_hint('p2_amplitude', value=1, min=0.01)
    model.set_param_hint('p3_amplitude', value=1, min=0.01)
    model.name = '3-gaussians'
    return model

## lmfit composite model used for fitting
def factory_two_asym_gaussians(add_bridge=False, p1_center=0.1, p2_center=0.9,
                               p1_sigma=0.03, p2_sigma=0.03):
    """Return a 2-Asym-Gaussians + (optional) bridge model that can fit data.

    The Asym-Gaussian function is :func:`asym_gaussian`.

    Arguments:
        add_bridge (bool): if True adds a bridge function between the two
            gaussian peaks. If False the model has only two Asym-Gaussians.

    The other arguments are initial values for the model parameters.

    Returns
        An `lmfit.Model` object with all the parameters already initialized.
    """
    peak1 = lmfit.model.Model(asym_gaussian, prefix='p1_')
    peak2 = lmfit.model.Model(asym_gaussian, prefix='p2_')
    model = peak1 + peak2

    model.set_param_hint('p1_center', value=p1_center, min=-1, max=2)
    #model.set_param_hint('p2_center', value=p2_center, min=-1, max=2)
    model.set_param_hint('p1_sigma1', value=p1_sigma, min=0.01, max=0.2)
    model.set_param_hint('p1_sigma2', value=p1_sigma, min=0.01, max=0.2)
    model.set_param_hint('p2_sigma1', value=p2_sigma, min=0.01, max=0.2)
    model.set_param_hint('p2_sigma2', value=p2_sigma, min=0.01, max=0.2)
    model.set_param_hint('p1_amplitude', value=1, min=0.01)
    model.set_param_hint('p2_amplitude', value=1, min=0.01)

    model.set_param_hint('pos_delta', value=0.3, min=0)
    model.set_param_hint('p2_center', min=-1, expr='p1_center + pos_delta')

    name = '2-asym-gaussians'

    if add_bridge:
        bridge = lmfit.model.Model(bridge_function, prefix='br_')
        model += bridge
        model.set_param_hint('br_amplitude', value=0.0001, min=0)
        model.set_params_hint('br_center1', min=-1, expr='p1_center')
        model.set_params_hint('br_center2', min=-1, expr='p2_center')
        model.set_params_hint('br_sigma1', min=0.01, expr='p1_sigma2')
        model.set_params_hint('br_sigma2', min=0.01, expr='p2_sigma1')
        name += '-bridge'
    model.name = name
    return model


##
# Classes to perform fit-related operations on multi-channel data
#
class FitterBase(object):
    """Base class for histogramming a dataset.

    To set weights assign attribute an array with same size as `data` to
    the attribute `.weights`.
    """
    def __init__(self, data):
        self.data = data
        self.weights = None
        self._hist_computed = False

    def histogram(self, **kwargs):
        """Compute the histogram of the data.

        All the kwargs are passed to `numpy.histogram`.
        """
        kwargs.update(density=False)
        if self.weights is not None:
            kwargs.update(weights=self.weights)
        hist_counts, bins = np.histogram(self.data, **kwargs)
        self._set_hist_data(hist_counts, bins)

    def _set_hist_data(self, hist_counts, bins):
        self.hist_bins = bins
        self.hist_binwidth = (bins[1] - bins[0])
        self.hist_axis = bins[:-1] + 0.5*self.hist_binwidth
        self.hist_counts = np.array(hist_counts)
        self.hist_pdf = np.array(hist_counts, dtype=np.float)
        self.hist_pdf /= self.hist_counts.sum(1)[:, np.newaxis]
        self.hist_pdf /= self.hist_binwidth
        self._hist_computed = True

    @property
    def x_axis(self):
        if not hasattr(self, '_x_axis'):
            self._x_axis = np.linspace(self.hist_axis[0],
                                       self.hist_axis[-1], 1000)
        return self._x_axis


class MultiFitter(FitterBase):
    """A class handling a list of 1-D datasets for histogramming, KDE, fitting.

    This class takes a list of 1-D arrays of samples (such as E values
    per burst). The list contains one 1-D array for each channel in
    a multispot experiment. In single-spot experiments the list contains only
    one array of samples.
    For each dataset in the list, this class compute histograms, KDEs and
    fits (both histogram fit and KDE maximum). The list of datasets is
    stored in the attribute `data_list`.
    The histograms can be fitted with an arbitrary model (lmfit.Model).
    From KDEs the peak position in a range can be estimated.

    Optionally weights can be assigned to each element in a dataset.
    To assign weights a user can assign the `.weights` attribute with a list
    of arrays; corresponding arrays in `.weights` and `.data_list` must have
    the same size.

    Alternatively a function returning the weights can be used. In this case,
    the method `.set_weights_func` allows to set the function to be called
    to generate weights.
    """
    def __init__(self, data_list, skip_ch=None):
        self.data_list = data_list
        self.ndata = len(data_list)
        self._weights = [None]*self.ndata
        self._hist_computed = False
        self._kde_computed = False
        self.skip_ch = []
        if skip_ch is not None:
            self.skip_ch = skip_ch

    @property
    def weights(self):
        return self._weights

    @weights.setter
    def weights(self, values):
        try:
            self._weights = [v for v in values]
        except TypeError:
            # values is not iterable
            self._weights = [values]*self.ndata

    def histogram(self, binwidth=0.03, bins=None, verbose=False, **kwargs):
        """Compute the histogram of the data for each channel.

        If `bins` is None, `binwidth` determines the bins array (saved in
        `self.hist_bins`). If `bins` is not None, `binwidth` is ignored and
        `self.hist_binwidth` is computed from `self.hist_bins`.

        The kwargs and `bins` are passed to `numpy.histogram`.
        """
        # Check if the same histogram is already computed
        if self._hist_computed:
            if bins is None and self.hist_binwidth == binwidth:
                return
            elif bins is not None and np.array_equal(self.hist_bins, bins):
                return

        if verbose:
            print(" - Computing histogram.")
        if bins is None:
            bins = np.arange(-0.2, 1.2, binwidth)
        kwargs.update(bins=bins, density=False)
        hist_counts = []
        for ich, (data, weights) in enumerate(zip(self.data_list, self.weights)):
            # NaN cause issues in some versions of numpy.histogram()
            valid = ~np.isnan(data)
            if ich in self.skip_ch:
                counts = np.zeros(bins.size - 1)
            else:
                if weights is not None:
                    kwargs.update(weights=weights[valid])
                counts, _ = np.histogram(data[valid], **kwargs)
            hist_counts.append(counts)
        self._set_hist_data(hist_counts, bins)

    def set_weights_func(self, weight_func, weight_kwargs=None):
        """Setup of the function returning the weights for each data-set.

        To compute the weights for each dataset the `weight_func` is called
        multiple times. Keys in `weight_kwargs` are arguments of
        `weight_func`. Values in `weight_kwargs` are either scalars, in which
        case they are passed to `weight_func`, or lists. When an argument
        is a list, only one element of the list is passed each time.

        Arguments:
            weight_func (function): function that returns the weights
            weight_kwargs (dict): keyword arguments to be passed to
                `weight_func`.
        """
        self.weights = []
        for i in range(self.ndata):
            weight_kw_i = {k: v[i] if isinstance(v, list) else v
                           for k, v in weight_kwargs.items()}
            self.weights.append(weight_func(**weight_kw_i))

    def fit_histogram(self, model=None, pdf=True, **fit_kwargs):
        """Fit the histogram of each channel using the same lmfit model.

        A list of `lmfit.Minimizer` is stored in `.fit_res`.
        The fitted values for all the parameters and all the channels are
        save in a Pandas DataFrame `.params`.

        Arguments:
            model (lmfit.Model object): lmfit Model with all the parameters
                already initialized used for fitting.
            pdf (bool): if True fit the normalized histogram (.hist_pdf)
                otherwise fit the raw counts (.hist_counts).
            fit_kwargs (dict): keyword arguments passed to `model().fit`.
        """
        if model is not None:
            self.model = model
        if not self._hist_computed:
            self.histogram()

        data_list = self.hist_pdf if pdf else self.hist_counts
        self.params = pd.DataFrame(index=range(self.ndata),
                                   columns=sorted(self.model.param_names))
        self.fit_res = []
        #init_params = copy.deepcopy(self.model.params)
        for i, data in enumerate(data_list):
            self.fit_res.append(
                self.model.fit(data, x=self.hist_axis, **fit_kwargs))
            self.params.iloc[i] = pd.Series(self.fit_res[-1].values)

    def calc_kde(self, bandwidth=0.03):
        """Compute the list of kde functions and save it in `.kde`.
        """
        if self._kde_computed and self.kde_bandwidth == bandwidth:
            return

        self.kde_bandwidth = bandwidth
        self.kde = []
        empty = []
        for ich, (data, weights_i) in enumerate(zip(self.data_list,
                                                    self.weights)):
            if ich in self.skip_ch or len(data) == 0:
                kde = empty
            else:
                kde = gf.gaussian_kde_w(data, bw_method=bandwidth,
                                        weights=weights_i)
            self.kde.append(kde)
        self._kde_computed = True

    def find_kde_max(self, x_kde, xmin=None, xmax=None):
        """Finds the peak position of kde functions between `xmin` and `xmax`.

        Results are saved in the list `.kde_max_pos`.
        """
        if not hasattr(self, 'kde'):
            self.calc_kde()
        self.kde_max_pos = np.zeros(self.ndata) * np.nan
        for ich, kde in enumerate(self.kde):
            if ich not in self.skip_ch or len(kde) == 0:
                self.kde_max_pos[ich] = find_max(x_kde, kde(x_kde),
                                                 xmin=xmin, xmax=xmax)


def plot_mfit(fitter, ich=0, residuals=False, ax=None, plot_kde=False,
              plot_model=True, return_fig=False, bins=None):
    """Plot data histogram and fitted model from a :class:`MultiFitter` object.

    Assumes data between 0 and 1 and a two peaks model with parameters
    `p1_center` and `p2_center`.

    Return
        A matplotlib figure object.
    """
    if not has_matplotlib:
        raise RuntimeError('Matplotlib not installed.')
    if ax is None:
        fig, ax = plt.subplots(figsize=(7, 4.5))
    else:
        fig = ax.figure

    if not fitter._hist_computed or bins is not None:
        fitter.histogram(bins=bins)
    ax.plot(fitter.hist_axis, fitter.hist_pdf[ich], '-o', alpha=0.5)
    num_bursts = fitter.data_list[ich].size
    ax.set_title('CH = %d #Bursts %d' % (ich, num_bursts))
    ax.set_xlim(-0.19, 1.19)
    ax.grid(True)
    red = '#E41A1C'

    if hasattr(fitter, 'fit_res') and plot_model:
        fit_res = fitter.fit_res[ich]
        x = fitter.x_axis
        ax.plot(x, fit_res.model.eval(x=x, **fit_res.values), 'k', alpha=0.8)
        if fit_res.model.components is not None:
            for component in fit_res.model.components:
                ax.plot(x, component.eval(x=x, **fit_res.values), '--k',
                        alpha=0.8)
        for param in fitter.params:
            if param.endswith('center'):
                ax.axvline(fitter.params[param][ich], ls='--', color=red)

        if residuals:
            ax.xaxis.set_ticklabels([])
            divider = make_axes_locatable(ax)
            ax_resid = divider.append_axes("bottom", size=1.2, pad=0.1,
                                           sharex=ax)
            ax_resid.plot(fitter.hist_axis,
                          fitter.hist_pdf[ich] - fit_res.best_fit)
            ax_resid.set_yticks(np.r_[-2:2:0.5])
            ax_resid.set_ylim(-0.49, 0.49)
            ax_resid.set_xlim(-0.2, 1.2)
    elif hasattr(fitter, 'kde') and plot_kde:
        kde = fitter.kde[ich]
        x = fitter.x_axis
        ax.plot(x, kde(x), color='gray', alpha=0.8)
        if hasattr(fitter, 'kde_max_pos'):
            ax.axvline(fitter.kde_max_pos[ich], ls='--', color=red)

    if return_fig:
        return fig



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014 Antonino Ingargiola <tritemio@gmail.com>
#
"""
This module provides functions to store and load background fit data
to and from an HDF5 file.

Background cache implementation
-------------------------------

Background caching only works when `bg_fun = bg.exp_fit` (MLE tail fit) and
assumes that `bg_ph_sel == Ph_sel('all')`.

Background estimation results are indentified by the `Data.calc_bg` arguments::

    time_s, tail_min_us, F_bg, error_metrics, fit_allph

These are serialized and used as group name under `/background`.
This group stores::

    bg

the arrays::

    Lim, Ph_p

and optionally the field `bg_auto_th_us0` used when `tail_min_us == 'auto'`.
The following `Data` attributes are computed (and not stored):

- bg_th_us_user: is tail_min_us when `tail_min_us != 'auto'`
- bg_auto_F_bg: is F_bg when `tail_min_us == 'auto'`

Finally the following `Data` attributes are fixed:

- bg_fun: fixed to bg.exp_fit, not saved but implied
- bg_ph_sel: fixed to Ph_sel('all'), not saved but implied

The following properties are not stored since they are compute on-fly every time:

- nperiods
- bg_mean

Attributes not saved nor restored (so far):

- Th_us: actual threshold to select the tail of the interphoton delays
  distribution. Dict of lists just like Data.bg.
- bg_err: metric for fit error estimation. Dict of lists just like Data.bg.
- bg_fun_name: equal to fun.__name__


Burst search caching
--------------------

Data.burst_search()
Input arguments: L, m, F, P, min_rate_cps, ph_sel, compact, index_allph, c

Other arguments: computefret=True, max_rate=False, dither=False,
                 pure_python=False, verbose=False, mute=False

Problem: the background estimation and periods influence burst search.
    In particular only Data.bg_from(ph_sel) is used. A viable approach is
    hashing Data.bg_from(ph_sel) and comparing it with the hash stored in the
    burst-search cache.

Attributes saved:
    m, L, ph_sel: scalar parameters

    Sets also:
        bg_corrected=False, leakage_corrected=False,
        dir_ex_corrected=False, dithering=False

Depending on input executes either:

    1)  _burst_search_rate()
    or
    2) _calc_T()
       _burst_search_TT()

Data._burst_search_rate()
Attributes saved:
    T, rate_th: per-spot values
    mburst: burst data

Data._calc_T():
Attributes saved:
    bg_bs, F, P: scalar parameters
    T, rate_th: per-spot means
    TT, FF, PP: list of arrays, per-spot and per background period

Data._burst_search_TT()
Attributes saved:
    mburst: burst data

Data._calc_burst_period():
Attributes saved:
    bp: list of arrays, uses Data.mburst and Data.Lim to compute Data.bp

Plan:
    - make signature from input args anche check is a matching group exist
    - if it exist and min_rate_cps is None, check if the background cache
      matches. It is good to check it also for min_rate_cps because background
      influences burst corrections.
    - if background hash matched can load background from cache: mbursts,
      T, rate_th
    - If min_rate_cps is None, call Data._calc_T() to recompute attributes:
      bg_bs, F, P, TT, FF, PP, T, rate_th. Assert that T and rate_th have not
      changed from previous step.
    - Call Data._calc_burst_period() to compute Data.bp.
    - Add to Data m, L, ph_sel (input parameters) and set
      bg_corrected=False, leakage_corrected=False,
      dir_ex_corrected=False, dithering=False
    - execute last part of burst-seach method that need to be refactored in a
      separate method.

"""

from __future__ import absolute_import
from builtins import range, zip

from pathlib import Path
import json
import numpy as np
import tables

from .utils.misc import pprint
from .ph_sel import Ph_sel
from .background import exp_fit


def bs_to_signature(L, m, F, P, min_rate_cps, ph_sel, compact, index_allph, c):
    return json.dumps(dict(L=L, m=m, F=F, P=P, min_rate_cps=min_rate_cps,
                           ph_sel=ph_sel, compact=compact,
                           index_allph=index_allph, c=c), sort_keys=True)


def bs_from_signature(string):
    return json.loads(string).items()


def bg_to_signature(d, time_s, tail_min_us, F_bg, error_metrics, fit_allph):
    params = dict(time_s=time_s, tail_min_us=tail_min_us, F_bg=F_bg,
                  error_metrics=error_metrics, fit_allph=fit_allph)
    if d.ALEX or 'PAX' in d.meas_type:
        params.update(offset=d.offset, d_on=d.D_ON, a_on=d.A_ON)
    return json.dumps(params, sort_keys=True)


def bg_from_signature(string):
    return {k: tuple(v) if isinstance(v, list) else v
            for k, v in json.loads(string).items()}


def _remove_cache_grp(h5file, group):
    """Remove `group` from `h5file`."""
    if group in h5file.root:
        h5file.remove_node(group, recursive=True)


def _remove_cache_bg(h5file):
    """Remove /background from `h5file`."""
    _remove_cache_grp(h5file, group='/background')


def _save_bg_data(d, bg, Lim, Ph_p, bg_calc_kwargs, h5file,
                  bg_auto_th_us0=None):
    """Save background data to HDF5 file."""
    # Save the bg data
    group_name = bg_to_signature(d, **bg_calc_kwargs)
    if _bg_is_cached(d, h5file, bg_calc_kwargs):
        h5file.remove_node('/background', group_name, recursive=True)

    bg_group = h5file.create_group('/background', group_name,
                                   createparents=True)

    for ph_sel in bg:
        h5file.create_array(bg_group, 'BG_%s' % str(ph_sel), obj=bg[ph_sel])
    h5file.create_array(bg_group, 'Lim', obj=Lim)
    h5file.create_array(bg_group, 'Ph_p', obj=Ph_p)
    if bg_calc_kwargs['tail_min_us'] == 'auto':
        assert bg_auto_th_us0 is not None
        h5file.create_array(bg_group, 'bg_auto_th_us0', obj=bg_auto_th_us0)


def _load_bg_data(d, bg_calc_kwargs, h5file):
    """Load background data from a HDF5 file."""
    group_name = bg_to_signature(d, **bg_calc_kwargs)
    if group_name not in h5file.root.background:
        msg = 'Group "%s" not found in the HDF5 file.' % group_name
        raise ValueError(msg)
    bg_auto_th_us0 = None
    bg_group = h5file.get_node('/background/', group_name)

    pprint('\n - Loading bakground data: ')
    bg = {}
    for node in bg_group._f_iter_nodes():
        if node._v_name.startswith('BG_'):
            ph_sel = Ph_sel.from_str(node._v_name[len('BG_'):])
            bg[ph_sel] = [np.asfarray(b) for b in node.read()]

    Lim = bg_group.Lim.read()
    Ph_p = bg_group.Ph_p.read()
    if 'bg_auto_th_us0' in bg_group:
        bg_auto_th_us0 = bg_group.bg_auto_th_us0.read()
    return bg, Lim, Ph_p, bg_auto_th_us0


def _bg_is_cached(d, h5file, bg_calc_kwargs):
    """Returns signature matches a group in /backgroung.
    """
    group_name = bg_to_signature(d, **bg_calc_kwargs)
    return ('background' in h5file.root and
            group_name in h5file.root.background)


def get_h5file(dx):
    datafile = Path(dx.fname)
    cachefile = datafile.with_name(datafile.stem + '_cache.hdf5')
    return tables.open_file(str(cachefile), mode='a')


def calc_bg_cache(dx, fun, time_s, tail_min_us, F_bg, error_metrics, fit_allph,
                  recompute=False):
    """Cached version of `.calc_bg()` method."""
    assert fun == exp_fit, 'Cache only supports the bg.exp_fit function.'
    assert error_metrics is None, 'Cache only support `error_metrics=None`.'
    bg_calc_kwargs = dict(time_s=time_s, tail_min_us=tail_min_us,
                          F_bg=F_bg, error_metrics=error_metrics,
                          fit_allph=fit_allph)
    h5file = get_h5file(dx)
    if _bg_is_cached(dx, h5file, bg_calc_kwargs) and not recompute:
        # Background found in cache. Load it.
        pprint(' * Loading BG rates from cache ... ')
        bg, Lim, Ph_p, bg_auto_th_us0 = _load_bg_data(dx, bg_calc_kwargs, h5file)

        bg_dict = dict(bg_fun=exp_fit, bg_ph_sel=Ph_sel('all'))     # fixed
        bg_dict.update(bg=bg, Lim=Lim, Ph_p=Ph_p, bg_time_s=time_s)
        if bg_calc_kwargs['tail_min_us'] == 'auto':
            bg_dict['bg_auto_F_bg'] = bg_calc_kwargs['F_bg']
            assert bg_auto_th_us0 is not None
            bg_dict['bg_auto_th_us0'] = bg_auto_th_us0
        else:
            bg_dict['bg_th_us_user'] = tail_min_us
        dx._clean_bg_data()
        dx.add(**bg_dict)
        pprint(' [DONE]\n')
    else:
        pprint(' * Computing BG rates:\n')
        dx.calc_bg(fun=fun, **bg_calc_kwargs)
        bg_auto_th_us0 = dx.get('bg_auto_th_us0', None)
        _save_bg_data(dx, dx.bg, dx.Lim, dx.Ph_p, bg_calc_kwargs, h5file,
                      bg_auto_th_us0=bg_auto_th_us0)
        pprint(' [DONE]\n')
    h5file.close()



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014-2016 Antonino Ingargiola <tritemio@gmail.com>
#
"""
This module provides functions to fit gaussian distributions and gaussian
distribution mixtures (2 components). These functions can be used directly,
or more often, in a typical FRETBursts workflow they are passed to higher
level methods like :meth:`fretbursts.burstlib.Data.fit_E_generic`.

Single Gaussian distribution fit:

    * :func:`gaussian_fit_hist`
    * :func:`gaussian_fit_cdf`
    * :func:`gaussian_fit_pdf`

For 2-Gaussians fit we have the following models:

    * :func:`two_gauss_mix_pdf`: *PDF of 2-components Gaussians mixture*
    * :func:`two_gauss_mix_ab`: *linear combination of 2 Gaussians*

Main functions for mixture of 2 Gaussian distribution fit:

    * :func:`two_gaussian_fit_hist` *histogram fit using `leastsq`*
    * :func:`two_gaussian_fit_hist_min` *histogram fit using `minimize`*
    * :func:`two_gaussian_fit_hist_min_ab` *the same but using _ab model*
    * :func:`two_gaussian_fit_cdf` *curve fit of the CDF*
    * :func:`two_gaussian_fit_EM` *Expectation-Maximization fit*
    * :func:`two_gaussian_fit_EM_b` *the same with boundaries*

Also, some functions to fit 2-D gaussian distributions and mixtures are
implemented but not thoroughly tested.

The reference documentation for **all** the functions follows.
"""

from __future__ import division, print_function, absolute_import

import numpy as np
import numpy.random as R
import scipy.optimize as O
import scipy.stats as S
from scipy.special import erf
from scipy.optimize import leastsq, minimize
import scipy.ndimage as ndi

#from scipy.stats import gaussian_kde
from .weighted_kde import gaussian_kde_w  # this version supports weights


def normpdf(x, mu=0, sigma=1.):
    """Return the normal pdf evaluated at `x`."""
    assert sigma > 0
    u = (x-mu)/sigma
    y = 1/(np.sqrt(2*np.pi)*sigma)*np.exp(-u*u/2)
    return y

##
# Single gaussian distribution
#

def gaussian_fit_curve(x, y, mu0=0, sigma0=1, a0=None, return_all=False,
        **kwargs):
    """Gaussian fit of curve (x,y).
    If a0 is None then only (mu,sigma) are fitted (to a gaussian density).
    `kwargs` are passed to the leastsq() function.

    If return_all=False then return only the fitted (mu,sigma) values
    If return_all=True (or full_output=True is passed to leastsq)
    then the full output of leastsq is returned.
    """
    if a0 is None:
        gauss_pdf = lambda x, m, s: np.exp(-((x-m)**2)/(2*s**2))/\
                                    (np.sqrt(2*np.pi)*s)
        err_fun = lambda p, x, y: gauss_pdf(x, *p) - y
        res = leastsq(err_fun, x0=[mu0, sigma0], args=(x, y), **kwargs)
    else:
        gauss_fun = lambda x, m, s, a: a*np.sign(s)*np.exp(-((x-m)**2)/(2*s**2))
        err_fun = lambda p, x, y: gauss_fun(x, *p) - y
        res = leastsq(err_fun, x0=[mu0, sigma0, a0], args=(x, y), **kwargs)

    if 'full_output' in kwargs:
        return_all = kwargs['full_output']
    mu, sigma = res[0][0], res[0][1]
    if return_all: return res
    return mu, sigma

def get_epdf(s, smooth=0, N=1000, smooth_pdf=False, smooth_cdf=True):
    """Compute the empirical PDF of the sample `s`.

    If smooth > 0 then apply a gaussian filter with sigma=smooth.
    N is the number of points for interpolation of the CDF on a uniform range.
    """
    ecdf = [np.sort(s), np.arange(0.5, s.size+0.5)*1./s.size]
    #ecdf = [np.sort(s), np.arange(s.size)*1./s.size]
    _x = np.linspace(s.min(), s.max(), N)
    ecdfi = [_x, np.interp(_x, ecdf[0], ecdf[1])]
    if smooth_cdf and smooth > 0:
        ecdfi[1] = ndi.filters.gaussian_filter1d(ecdfi[1], sigma=smooth)
    epdf = [ecdfi[0][:-1], np.diff(ecdfi[1])/np.diff(ecdfi[0])]
    if smooth_pdf and smooth > 0:
        epdf[1] = ndi.filters.gaussian_filter1d(epdf[1], sigma=smooth)
    return epdf

def gaussian_fit_pdf(s, mu0=0, sigma0=1, a0=1, return_all=False,
        leastsq_kwargs={}, **kwargs):
    """Gaussian fit of samples s using a fit to the empirical PDF.
    If a0 is None then only (mu,sigma) are fitted (to a gaussian density).
    `kwargs` are passed to get_epdf().
    If return_all=False then return only the fitted (mu,sigma) values
    If return_all=True (or full_output=True is passed to leastsq)
    then the full output of leastsq and the PDF curve is returned.
    """
    ## Empirical PDF
    epdf = get_epdf(s, **kwargs)

    res = gaussian_fit_curve(epdf[0], epdf[1], mu0, sigma0, a0, return_all,
            **leastsq_kwargs)
    if return_all: return res, epdf
    return res

def gaussian_fit_hist(s, mu0=0, sigma0=1, a0=None, bins=np.r_[-0.5:1.5:0.001],
        return_all=False, leastsq_kwargs={}, weights=None, **kwargs):
    """Gaussian fit of samples s fitting the hist to a Gaussian function.
    If a0 is None then only (mu,sigma) are fitted (to a gaussian density).
    kwargs are passed to the histogram function.
    If return_all=False then return only the fitted (mu,sigma) values
    If return_all=True (or full_output=True is passed to leastsq)
    then the full output of leastsq and the histogram is returned.
    `weights` optional weights for the histogram.
    """
    histogram_kwargs = dict(bins=bins, density=True, weights=weights)
    histogram_kwargs.update(**kwargs)
    H = np.histogram(s, **histogram_kwargs)
    x, y = 0.5*(H[1][:-1] + H[1][1:]), H[0]
    #bar(H[1][:-1], H[0], H[1][1]-H[1][0], alpha=0.3)

    res = gaussian_fit_curve(x, y, mu0, sigma0, a0, return_all,
            **leastsq_kwargs)
    if return_all: return res, H, x, y
    return res

def gaussian_fit_cdf(s, mu0=0, sigma0=1, return_all=False, **leastsq_kwargs):
    """Gaussian fit of samples s fitting the empirical CDF.
    Additional kwargs are passed to the leastsq() function.
    If return_all=False then return only the fitted (mu,sigma) values
    If return_all=True (or full_output=True is passed to leastsq)
    then the full output of leastsq and the histogram is returned.
    """
    ## Empirical CDF
    ecdf = [np.sort(s), np.arange(0.5, s.size+0.5)*1./s.size]

    ## Analytical Gaussian CDF
    gauss_cdf = lambda x, mu, sigma: 0.5*(1+erf((x-mu)/(np.sqrt(2)*sigma)))

    ## Fitting the empirical CDF
    err_func = lambda p, x, y: y - gauss_cdf(x, p[0], p[1])
    res = leastsq(err_func, x0=[mu0, sigma0], args=(ecdf[0], ecdf[1]),
            **leastsq_kwargs)
    if return_all: return res, ecdf
    return res[0]

def gaussian_fit_ml(s, mu_sigma_guess=[0.5, 1]):
    """Gaussian fit of samples s using the Maximum Likelihood (ML method).
    Didactical, since scipy.stats.norm.fit() implements the same method.
    """
    n = s.size
    ## Log-likelihood (to be maximized)
    log_l = lambda mu, sig: -n/2.*np.log(sig**2) - \
                             1./(2*sig**2)*np.sum((s-mu)**2)

    ## Function to be minimized
    min_fun = lambda p: -log_l(p[0], p[1])

    res = O.minimize(min_fun, [0, 0.5], method='powell',
                     options={'xtol': 1e-6, 'disp': True, 'maxiter': 1e9})

    print(res)
    mu, sigma = res['x']
    return mu, sigma

##
# Two-component gaussian mixtures
#

def two_gauss_mix_pdf(x, p):
    """PDF for the distribution of a mixture of two Gaussians."""
    mu1, sig1, mu2, sig2, a = p
    return a*normpdf(x, mu1, sig1) + (1-a)*normpdf(x, mu2, sig2)

def two_gauss_mix_ab(x, p):
    """Mixture of two Gaussians with no area constrain."""
    mu1, sig1, a1, mu2, sig2, a2 = p
    return a1*normpdf(x, mu1, sig1) + a2*normpdf(x, mu2, sig2)

def reorder_parameters(p):
    """Reorder 2-gauss mix params to have the 1st component with smaller mean.
    """
    if p[0] > p[2]:
        p = p[np.array([2, 3, 0, 1, 4])]  # swap (mu1, sig1) with (mu2, sig2)
        p[4] = 1 - p[4]                   # "swap" the alpha of the mixture
    return p

def reorder_parameters_ab(p):
    """Reorder 2-gauss mix params to have the 1st component with smaller mean.
    """
    if p[0] > p[3]:
        p = p[np.array([3, 4, 5, 0, 1, 2])]
    return p


def bound_check(val, bounds):
    """Returns `val` clipped inside the interval `bounds`."""
    if bounds[0] is not None and val < bounds[0]:
        val = bounds[0]
    if bounds[1] is not None and val > bounds[1]:
        val = bounds[1]
    return val

def two_gaussian_fit_curve(x, y, p0, return_all=False, verbose=False, **kwargs):
    """Fit a 2-gaussian mixture to the (x,y) curve.
    `kwargs` are passed to the leastsq() function.

    If return_all=False then return only the fitted paramaters
    If return_all=True then the full output of leastsq is returned.
    """
    if kwargs['method'] == 'leastsq':
        kwargs.pop('method')
        kwargs.pop('bounds')
        def err_func(p, x, y):
            return (y - two_gauss_mix_pdf(x, p))
        res = leastsq(err_func, x0=p0, args=(x, y), **kwargs)
        p = res[0]
    else:
        def err_func(p, x, y):
            return ((y - two_gauss_mix_pdf(x, p))**2).sum()
        res = minimize(err_func, x0=p0, args=(x, y), **kwargs)
        p = res.x

    if verbose:
        print(res, '\n')
    if return_all: return res
    return reorder_parameters(p)

def two_gaussian_fit_KDE_curve(s, p0=[0, 0.1, 0.6, 0.1, 0.5], weights=None,
                               bandwidth=0.05, x_pdf=None, debug=False,
                               method='SLSQP', bounds=None,
                               verbose=False, **kde_kwargs):
    """Fit sample `s` with two gaussians using a KDE pdf approximation.

    The 2-gaussian pdf is then curve-fitted to the KDE pdf.

    Arguments:
        s (array): population of samples to be fitted
        p0 (sequence-like): initial parameters [mu0, sig0, mu1, sig1, a]
        bandwidth (float): bandwidth for the KDE algorithm
        method (string): fit method, can be 'leastsq' or one of the methods
            accepted by scipy `minimize()`
        bounds (None or 5-element list): if not None, each element is a
            (min,max) pair of bounds for the corresponding parameter. This
            argument can be used only with L-BFGS-B, TNC or SLSQP methods.
            If bounds are used, parameters cannot be fixed
        x_pdf (array): array on which the KDE PDF is evaluated and curve-fitted
        weights (array): optional weigths, same size as `s` (for ex.
            1/sigma^2 ~ nt).
        debug (bool): if True perfoms more tests and print more info.

    Additional kwargs are passed to scipy.stats.gaussian_kde().

    Returns:
        Array of parameters for the 2-gaussians (5 elements)
    """
    if x_pdf is None: x_pdf = np.linspace(s.min(), s.max(), 1000)

    ## Scikit-learn KDE estimation
    #kde_skl = KernelDensity(bandwidth=bandwidth, **kde_kwargs)
    #kde_skl.fit(x)[:, np.newaxis])
    ## score_samples() returns the log-likelihood of the samples
    #log_pdf = kde_skl.score_samples(x_pdf)[:, np.newaxis])
    #kde_pdf = np.exp(log_pdf)

    ## Weighted KDE estimation
    kde = gaussian_kde_w(s, bw_method=bandwidth, weights=weights)
    kde_pdf = kde.evaluate(x_pdf)

    p = two_gaussian_fit_curve(x_pdf, kde_pdf, p0=p0, method=method,
                               bounds=bounds, verbose=verbose)
    return p


def two_gaussian_fit_EM_b(s, p0=[0, 0.1, 0.6, 0.1, 0.5], weights=None,
                          bounds=[(None, None,)]*5,
                          max_iter=300, ptol=1e-4, debug=False):
    """
    Fit the sample s with two gaussians using Expectation Maximization.

    This version allows setting boundaries for each parameter.

    Arguments:
        s (array): population of samples to be fitted
        p0 (sequence-like): initial parameters [mu0, sig0, mu1, sig1, a]
        bound (tuple of pairs): sequence of (min, max) values that constrain
            the parameters. If min or max are None, no boundary is set.
        ptol (float): convergence condition. Relative max variation of any
            parameter.
        max_iter (int): max number of iteration in case of non convergence.
        weights (array): optional weigths, same size as `s` (for ex.
            1/sigma^2 ~ nt).

    Returns:
        Array of parameters for the 2-gaussians (5 elements)
    """
    assert np.size(p0) == 5
    if weights is None: weights = np.ones(s.size)
    assert weights.size == s.size
    weights *= (1.*weights.size)/weights.sum() # Normalize to (#samples)
    #weights /= weights.sum()  # Normalize to 1
    if debug: assert np.abs(weights.sum() - s.size) < 1e-6
    bounds_mu = [bounds[0], bounds[2]]
    bounds_sig = [bounds[1], bounds[3]]
    bounds_pi0 = bounds[4]

    # Initial guess of parameters and initializations
    mu = np.array([p0[0], p0[2]])
    sig = np.array([p0[1], p0[3]])
    pi_ = np.array([p0[4], 1-p0[4]])

    gamma = np.zeros((2, s.size))
    N_ = np.zeros(2)
    p_new = np.array(p0)

    # EM loop
    counter = 0
    stop_iter, converged = False, False
    while not stop_iter:
        # Compute the responsibility func. (gamma) and the new parameters
        for k in [0, 1]:
            gamma[k, :] = weights*pi_[k]*normpdf(s, mu[k], sig[k]) / \
                            two_gauss_mix_pdf(s, p_new)
            N_[k] = gamma[k, :].sum()
            mu[k] = np.sum(gamma[k]*s)/N_[k]
            mu[k] = bound_check(mu[k], bounds_mu[k])
            sig[k] = np.sqrt( np.sum(gamma[k]*(s-mu[k])**2)/N_[k] )
            sig[k] = bound_check(sig[k], bounds_sig[k])
            if k < 1:
                pi_[k] = N_[k]/s.size
                pi_[k] = bound_check(pi_[k], bounds_pi0)
            else:
                pi_[k] = 1 - pi_[0]
        p_old = p_new
        p_new = np.array([mu[0], sig[0], mu[1], sig[1], pi_[0]])
        if debug:
            assert np.abs(N_.sum() - s.size)/float(s.size) < 1e-6
            assert np.abs(pi_.sum() - 1) < 1e-6

        # Convergence check
        counter += 1
        relative_delta = np.abs(p_new - p_old)/p_new
        converged = relative_delta.max() < ptol
        stop_iter = converged or (counter >= max_iter)

    if debug:
        print("Iterations: ", counter)
    if not converged:
        print("WARNING: Not converged, max iteration (%d) reached." % max_iter)
    return reorder_parameters(p_new)

def two_gaussian_fit_EM(s, p0=[0, 0.1, 0.6, 0.1, 0.5], max_iter=300, ptol=1e-4,
                        fix_mu=[0, 0], fix_sig=[0, 0], debug=False,
                        weights=None):
    """
    Fit the sample s with two gaussians using Expectation Maximization.

    This vesion allows to optionally fix mean or std. dev. of each component.

    Arguments:
        s (array): population of samples to be fitted
        p0 (sequence-like): initial parameters [mu0, sig0, mu1, sig1, a]
        bound (tuple of pairs): sequence of (min, max) values that constrain
            the parameters. If min or max are None, no boundary is set.
        ptol (float): convergence condition. Relative max variation of any
            parameter.
        max_iter (int): max number of iteration in case of non convergence.
        weights (array): optional weigths, same size as `s` (for ex.
            1/sigma^2 ~ nt).

    Returns:
        Array of parameters for the 2-gaussians (5 elements)
    """
    assert np.size(p0) == 5
    if weights is None: weights = np.ones(s.size)
    assert weights.size == s.size
    weights *= (1.*weights.size)/weights.sum() # Normalize to (#samples)
    #weights /= weights.sum()  # Normalize to 1
    if debug: assert np.abs(weights.sum() - s.size) < 1e-6

    # Initial guess of parameters and initializations
    mu = np.array([p0[0], p0[2]])
    sig = np.array([p0[1], p0[3]])
    pi_ = np.array([p0[4], 1-p0[4]])

    gamma = np.zeros((2, s.size))
    N_ = np.zeros(2)
    p_new = np.array(p0)

    # EM loop
    counter = 0
    stop_iter, converged = False, False
    while not stop_iter:
        # Compute the responsibility func. (gamma) and the new parameters
        for k in [0, 1]:
            gamma[k, :] = weights*pi_[k]*normpdf(s, mu[k], sig[k]) / \
                    two_gauss_mix_pdf(s, p_new)
            ## Uncomment for SCHEME2
            #gamma[k, :] = pi_[k]*normpdf(s, mu[k], sig[k]) / \
            #        two_gauss_mix_pdf(s, p_new)
            N_[k] = gamma[k, :].sum()
            if not fix_mu[k]:
                mu[k] = np.sum(gamma[k]*s)/N_[k]
                ## Uncomment for SCHEME2
                #mu[k] = np.sum(weights*gamma[k]*s)/N_[k]
            if not fix_sig[k]:
                sig[k] = np.sqrt( np.sum(gamma[k]*(s-mu[k])**2)/N_[k] )
            pi_[k] = N_[k]/s.size
        p_old = p_new
        p_new = np.array([mu[0], sig[0], mu[1], sig[1], pi_[0]])
        if debug:
            assert np.abs(N_.sum() - s.size)/float(s.size) < 1e-6
            assert np.abs(pi_.sum() - 1) < 1e-6

        # Convergence check
        counter += 1
        fixed = np.concatenate([fix_mu, fix_sig, [0]]).astype(bool)
        relative_delta = np.abs(p_new[~fixed] - p_old[~fixed])/p_new[~fixed]
        converged = relative_delta.max() < ptol
        stop_iter = converged or (counter >= max_iter)

    if debug:
        print("Iterations: ", counter)
    if not converged:
        print("WARNING: Not converged, max iteration (%d) reached." % max_iter)
    return reorder_parameters(p_new)

def two_gaussian_fit_hist(s, bins=np.r_[-0.5:1.5:0.001], weights=None,
        p0=[0.2,1,0.8,1,0.3], fix_mu=[0,0], fix_sig=[0,0], fix_a=False):
    """Fit the sample s with 2-gaussian mixture (histogram fit).

    Uses scipy.optimize.leastsq function. Parameters can be fixed but
    cannot be constrained in an interval.

    Arguments:
        s (array): population of samples to be fitted
        p0 (5-element list or array): initial guess or parameters
        bins (int or array): bins passed to `np.histogram()`
        weights (array): optional weights passed to `np.histogram()`
        fix_a (tuple of bools): Whether to fix the amplitude of the gaussians
        fix_mu (tuple of bools): Whether to fix the mean of the gaussians
        fix_sig (tuple of bools): Whether to fix the sigma of the gaussians

    Returns:
        Array of parameters for the 2-gaussians (5 elements)
    """
    assert np.size(p0) == 5
    fix = np.array([fix_mu[0], fix_sig[0], fix_mu[1], fix_sig[1], fix_a],
                dtype=bool)
    p0 = np.array(p0)
    p0_free = p0[-fix]
    p0_fix = p0[fix]

    H = np.histogram(s, bins=bins, weights=weights, density=True)
    x, y = 0.5*(H[1][:-1] + H[1][1:]), H[0]
    assert x.size == y.size

    ## Fitting
    def err_func(p, x, y, fix, p_fix, p_complete):
        p_complete[-fix] = p
        p_complete[fix] = p_fix
        return y - two_gauss_mix_pdf(x, p_complete)

    p_complete = np.zeros(5)
    p, v = leastsq(err_func, x0=p0_free, args=(x, y, fix, p0_fix, p_complete))
    p_new = np.zeros(5)
    p_new[-fix] = p
    p_new[fix] = p0_fix
    return reorder_parameters(p_new)


def two_gaussian_fit_hist_min(s, bounds=None, method='L-BFGS-B',
        bins=np.r_[-0.5:1.5:0.001], weights=None,  p0=[0.2,1,0.8,1,0.3],
        fix_mu=[0,0], fix_sig=[0,0], fix_a=False, verbose=False):
    """Fit the sample `s` with 2-gaussian mixture (histogram fit). [Bounded]

    Uses scipy.optimize.minimize allowing constrained minimization.

    Arguments:
        s (array): population of samples to be fitted
        method (string): one of the methods accepted by scipy `minimize()`
        bounds (None or 5-element list): if not None, each element is a
            (min,max) pair of bounds for the corresponding parameter. This
            argument can be used only with L-BFGS-B, TNC or SLSQP methods.
            If bounds are used, parameters cannot be fixed
        p0 (5-element list or array): initial guess or parameters
        bins (int or array): bins passed to `np.histogram()`
        weights (array): optional weights passed to `np.histogram()`
        fix_a (tuple of bools): Whether to fix the amplitude of the gaussians
        fix_mu (tuple of bools): Whether to fix the mean of the gaussians
        fix_sig (tuple of bools): Whether to fix the sigma of the gaussians
        verbose (boolean): allows printing fit information

    Returns:
        Array of parameters for the 2-gaussians (5 elements)
    """
    assert np.size(p0) == 5
    fix = np.array([fix_mu[0], fix_sig[0], fix_mu[1], fix_sig[1], fix_a],
                dtype=bool)
    p0 = np.array(p0)
    p0_free = p0[-fix]
    p0_fix = p0[fix]

    H = np.histogram(s, bins=bins, weights=weights, density=True)
    x, y = 0.5*(H[1][:-1] + H[1][1:]), H[0]
    assert x.size == y.size

    ## Fitting
    def err_func(p, x, y, fix, p_fix, p_complete):
        p_complete[-fix] = p
        p_complete[fix] = p_fix
        return ((y - two_gauss_mix_pdf(x, p_complete))**2).sum()

    p_complete = np.zeros(5)
    res = minimize(err_func, x0=p0_free, args=(x, y, fix, p0_fix, p_complete),
                    method=method, bounds=bounds)
    if verbose: print(res)
    p_new = np.zeros(5)
    p_new[-fix] = res.x
    p_new[fix] = p0_fix
    return reorder_parameters(p_new)

def two_gaussian_fit_hist_min_ab(s, bounds=None, method='L-BFGS-B',
        bins=np.r_[-0.5:1.5:0.001], weights=None,  p0=[0.2,1,0.8,1,0.3],
        fix_mu=[0,0], fix_sig=[0,0], fix_a=[0,0], verbose=False):
    """Histogram fit of sample `s` with 2-gaussian functions.

    Uses scipy.optimize.minimize allowing constrained minimization. Also
    each parameter can be fixed.

    The order of the parameters is: mu1, sigma1, a1, mu2, sigma2, a2.

    Arguments:
        s (array): population of samples to be fitted
        method (string): one of the methods accepted by scipy `minimize()`
        bounds (None or 6-element list): if not None, each element is a
            (min,max) pair of bounds for the corresponding parameter. This
            argument can be used only with L-BFGS-B, TNC or SLSQP methods.
            If bounds are used, parameters cannot be fixed
        p0 (6-element list or array): initial guess or parameters
        bins (int or array): bins passed to `np.histogram()`
        weights (array): optional weights passed to `np.histogram()`
        fix_a (tuple of bools): Whether to fix the amplitude of the gaussians
        fix_mu (tuple of bools): Whether to fix the mean of the gaussians
        fix_sig (tuple of bools): Whether to fix the sigma of the gaussians
        verbose (boolean): allows printing fit information

    Returns:
        Array of parameters for the 2-gaussians (6 elements)
    """
    nparams = 6
    assert np.size(p0) == nparams
    fix = np.array([fix_mu[0], fix_sig[0], fix_a[0],
                    fix_mu[1], fix_sig[1], fix_a[1]], dtype=bool)
    p0 = np.asarray(p0)
    p0_free = p0[-fix]
    p0_fix = p0[fix]

    counts, bins = np.histogram(s, bins=bins, weights=weights, density=True)
    x = bins[:-1] + 0.5*(bins[1] - bins[0])
    y = counts
    assert x.size == y.size

    ## Fitting
    def err_func(p, x, y, fix, p_fix, p_complete):
        p_complete[-fix] = p
        p_complete[fix] = p_fix
        return ((y - two_gauss_mix_ab(x, p_complete))**2).sum()

    p_complete = np.zeros(nparams)
    res = minimize(err_func, x0=p0_free, args=(x, y, fix, p0_fix, p_complete),
                    method=method, bounds=bounds)
    if verbose: print(res)
    p_new = np.zeros(nparams)
    p_new[-fix] = res.x
    p_new[fix] = p0_fix
    return reorder_parameters_ab(p_new)

def two_gaussian_fit_cdf(s, p0=[0., .05, .6, .1, .5],
                         fix_mu=[0, 0], fix_sig=[0, 0]):
    """Fit the sample s with two gaussians using a CDF fit.

    Curve fit 2-gauss mixture Cumulative Distribution Function (CDF) to the
    empirical CDF for sample `s`.

    Note that with a CDF fit no weighting is possible.

    Arguments:
        s (array): population of samples to be fitted
        p0 (5-element list or array): initial guess or parameters
        fix_mu (tuple of bools): Whether to fix the mean of the gaussians
        fix_sig (tuple of bools): Whether to fix the sigma of the gaussians

    Returns:
        Array of parameters for the 2-gaussians (5 elements)
    """
    assert np.size(p0) == 5
    fix = np.array([fix_mu[0], fix_sig[0], fix_mu[1], fix_sig[1], 0],
                    dtype=bool)
    p0 = np.array(p0)
    p0_free = p0[-fix]
    p0_fix = p0[fix]

    ## Empirical CDF
    ecdf = [np.sort(s), np.arange(0.5, s.size+0.5)*1./s.size]
    x, y = ecdf

    ## Analytical gaussian CDF
    gauss_cdf = lambda x, mu, sigma: 0.5*(1+erf((x-mu)/(np.sqrt(2)*sigma)))
    def two_gauss_mix_cdf(x, p):
        return p[4]*gauss_cdf(x, p[0], p[1]) + (1-p[4])*gauss_cdf(x, p[2], p[3])

    ## Fitting the empirical CDF
    def err_func(p, x, y, fix, p_fix, p_complete):
        p_complete[-fix] = p
        p_complete[fix] = p_fix
        return y - two_gauss_mix_cdf(x, p_complete)

    p_complete = np.zeros(5)
    p, v = leastsq(err_func, x0=p0_free, args=(x, y, fix, p0_fix, p_complete))
    p_new = np.zeros(5)
    p_new[-fix] = p
    p_new[fix] = p0_fix
    return reorder_parameters(p_new)

def test_two_gauss():
    m01 = 0.
    m02 = 0.6
    s01 = 0.05
    s02 = 0.1
    alpha = 0.
    p_real = [m01, s01, m02, s02, alpha]

    N = 500
    si1 = round(alpha*N)
    si2 = round((1-alpha)*N)
    s1 = R.normal(size=si1, loc=m01, scale=s01)
    s2 = R.normal(size=si2, loc=m02, scale=s02)
    s = np.r_[s1,s2]

    pc = two_gaussian_fit_cdf(s, fix_mu=[1,0], p0=[-0.01,0.05,0.5,0.2,0.4])
    ph = two_gaussian_fit_hist(s, fix_mu=[1,0], p0=[-0.01,0.05,0.5,0.2,0.4])
    pe = two_gaussian_fit_EM(s, fix_mu=[1,0], p0=[-0.01,0.05,0.5,0.2,0.4])

    hist(s, bins=40, normed=True)

    x = np.r_[s.min()-1:s.max()+1:200j]
    plot(x, a*normpdf(x,mu1,sig1), lw=2)
    plot(x, (1-a)*normpdf(x,mu2,sig2), lw=2)
    plot(x, two_gauss_mix_pdf(x, p0), lw=2)

    axvline(m01, lw=2, color='k', alpha=0.3)
    axvline(m02, lw=2, color='gray', alpha=0.3)
    axvline(mu1, lw=2, ls='--', color='k', alpha=0.3)
    axvline(mu2, lw=2, ls='--', color='gray', alpha=0.3)
    axvline(mu1h, lw=2, ls='--', color='r', alpha=0.3)
    axvline(mu2h, lw=2, ls='--', color='r', alpha=0.3)

def compare_two_gauss():
    m01 = 0.
    m02 = 0.5
    s01 = 0.08
    s02 = 0.15
    alpha = 0.7
    p_real = [m01, s01, m02, s02, alpha]

    N = 1000
    si1 = round(N*alpha)
    si2 = round((1-alpha)*N)

    p0 = [-0.01,0.05,0.6,0.2,0.4]
    fix_mu = [0,0]

    n = 500
    PC, PH, PE = np.zeros((n,5)), np.zeros((n,5)), np.zeros((n,5))
    for i in xrange(n):
        s1 = R.normal(size=si1, loc=m01, scale=s01)
        s2 = R.normal(size=si2, loc=m02, scale=s02)
        s = np.r_[s1,s2]

        pc = two_gaussian_fit_cdf(s, fix_mu=fix_mu, p0=p0)
        ph = two_gaussian_fit_hist(s, fix_mu=fix_mu, p0=p0)
        pe = two_gaussian_fit_EM(s, fix_mu=fix_mu, p0=p0)

        PC[i], PH[i], PE[i] = pc, ph, pe

    Label = ['Mu1', 'Sig1', 'Mu2', 'Sig2', 'Alpha']
    ftype = 'png'

    for i in range(5):
        figure()
        title(Label[i])
        vmin = min([PC[:,i].min(), PH[:,i].min(), PE[:,i].min()])
        vmax = max([PC[:,i].max(), PH[:,i].max(), PE[:,i].max()])
        b = np.r_[vmin:vmax:80j]
        if vmax == vmin: b = np.r_[vmin-.1:vmax+.1:200j]
        hist(PC[:,i], bins=b, alpha=0.3, label='CDF')
        hist(PH[:,i], bins=b, alpha=0.3, label='Hist')
        hist(PE[:,i], bins=b, alpha=0.3, label='EM')
        legend(loc='best')
        axvline(p_real[i], color='k', lw=2)
        #savefig('Two-gaussian Fit Comp - %s.png' % Label[i])

def gaussian2d_fit(sx, sy, guess=[0.5,1]):
    """2D-Gaussian fit of samples S using a fit to the empirical CDF."""
    assert sx.size == sy.size

    ## Empirical CDF
    ecdfx = [np.sort(sx), np.arange(0.5,sx.size+0.5)*1./sx.size]
    ecdfy = [np.sort(sy), np.arange(0.5,sy.size+0.5)*1./sy.size]

    ## Analytical gaussian CDF
    gauss_cdf = lambda x, mu, sigma: 0.5*(1+erf((x-mu)/(np.sqrt(2)*sigma)))

    ## Fitting the empirical CDF
    fitfunc = lambda p, x: gauss_cdf(x, p[0], p[1])
    errfunc = lambda p, x, y: fitfunc(p, x) - y
    px,v = leastsq(errfunc, x0=guess, args=(ecdfx[0],ecdfx[1]))
    py,v = leastsq(errfunc, x0=guess, args=(ecdfy[0],ecdfy[1]))
    print("2D Gaussian CDF fit", px, py)

    mux, sigmax = px[0], px[1]
    muy, sigmay = py[0], py[1]
    return mux, sigmax, muy, sigmay

def test_gaussian2d_fit():
    mx0 = 0.1
    my0 = 0.9
    sigx0 = 0.4
    sigy0 = 0.25

    Size = 500
    sx = R.normal(size=Size, loc=mx0, scale=sigx0)
    sy = R.normal(size=Size, loc=my0, scale=sigy0)

    mux, sigmax, muy, sigmay = gaussian2d_fit(sx, sy)

    plot(sx, sy, 'o', alpha=0.2, mew=0)

    X,Y = np.mgrid[sx.min()-1:sx.max()+1:200j, sy.min()-1:sy.max()+1:200j]

    def gauss2d(X,Y, mx, my, sigx, sigy):
        return np.exp(-((X-mx)**2)/(2*sigx**2))*np.exp(-((Y-my)**2)/(2*sigy**2))

    contour(X,Y,gauss2d(X,Y,mux,muy,sigmax,sigmay))

    plot(mx0,my0, 'ok', mew=0, ms=10)
    plot(mux,muy, 'x', mew=2, ms=10, color='green')


def two_gaussian2d_fit(sx, sy, guess=[0.5,1]):
    """2D-Gaussian fit of samples S using a fit to the empirical CDF."""
    ## UNFINISHED (I have 2 alphas unp.sign the xy projections)
    assert sx.size == sy.size

    ## Empirical CDF
    ecdfx = [np.sort(sx), np.arange(0.5,sx.size+0.5)*1./sx.size]
    ecdfy = [np.sort(sy), np.arange(0.5,sy.size+0.5)*1./sy.size]

    ## Analytical gaussian CDF
    gauss_cdf = lambda x, mu, sigma: 0.5*(1+erf((x-mu)/(np.sqrt(2)*sigma)))

    gauss2d_cdf = lambda X,Y,mx,sx,my,sy: gauss_cdf(X,mx,sx)*gauss_cdf(Y,my,sy)

    two_cdf = lambda x, m1, s1, m2, s2, a:\
        a*gauss_cdf(x,m1,s1)+(1-a)*gauss_cdf(x,m2,s2)

    two2d_cdf = lambda X,Y, mx1, sx1, mx2, sx2, my1, sy1, my2, sy2, a:\
        a*gauss2d_cdf(X,Y,mx1,sx1,my1,sy1)+(1-a)*gauss_cdf(X,Y,mx2,sx2,my2,sy2)

    ## Fitting the empirical CDF
    fitfunc = lambda p, x: two_cdf(x, *p)
    errfunc = lambda p, x, y: fitfunc(p, x) - y
    fitfunc2d = lambda p, X,Y: two2d_cdf(X,Y, *p)
    errfunc2d = lambda p, X,Y,Z: fitfunc2d(p, X,Y) - Z

    px,v = leastsq(errfunc, x0=guess, args=(ecdfx[0],ecdfx[1]))
    py,v = leastsq(errfunc, x0=guess, args=(ecdfy[0],ecdfy[1]))
    print("2D Two-Gaussians CDF fit", px, py)

    mux1, sigmax1, mux2, sigmax2, alphax = px
    muy1, sigmay1, muy2, sigmay2, alphay = py
    return mu1, sigma1, mu2, sigma2, alpha

def test_gaussian_fit():
    m0 = 0.1
    s0 = 0.4
    size = 500

    s = R.normal(size=size, loc=m0, scale=s0)
    #s = s[s<0.4]
    mu, sig = gaussian_fit(s)
    mu1, sig1 = S.norm.fit(s)
    mu2, sig2 = gaussian_fit_ml(s)

    print("ECDF ", mu, sig)
    print("ML         ", mu1, sig1)
    print("ML (manual)", mu2, sig2)

    H = np.histogram(s, bins=20, density=True)
    h = H[0]
    bw = H[1][1] - H[1][0]
    #bins_c = H[1][:-1]+0.5*bw
    bar(H[1][:-1], H[0], bw, alpha=0.3)

    x = np.r_[s.min()-1:s.max()+1:200j]
    plot(x, normpdf(x,m0,s0), lw=2, color='grey')
    plot(x, normpdf(x,mu,sig), lw=2, color='r', alpha=0.5)
    plot(x, normpdf(x,mu1,sig1), lw=2, color='b', alpha=0.5)

if __name__ == '__main__':
    #compare_two_gauss()
    #test_gaussian2d_fit()
    #test_gaussian_fit()
    #show()
    pass
